import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.distributions as dist
from torch import optim
import math
import numpy as np
from numpy import prod, sqrt
from pyro.distributions.zero_inflated import ZeroInflatedNegativeBinomial
from torch.utils.data import DataLoader
from sklearn.cluster import KMeans, DBSCAN
from sklearn.neighbors import NearestNeighbors
from dance.utils import SimpleIndexDataset
import os
import scipy.sparse as sp
import pandas as pd

def get_mean(d, K=100):
    """
    Extract the `mean` parameter for given distribution.
    If attribute not available, estimate from samples.
    """
    try:
        mean = d.mean
    except NotImplementedError:
        samples = d.rsample(torch.Size([K]))
        mean = samples.mean(0)
    return mean

def kl_divergence(d1, d2, K=100):
    """Computes closed-form KL if available, else computes a MC estimate."""
    if (type(d1), type(d2)) in torch.distributions.kl._KL_REGISTRY:
        return torch.distributions.kl_divergence(d1, d2)
    else:
        samples = d1.rsample(torch.Size([K]))
        return (d1.log_prob(samples) - d2.log_prob(samples)).mean(0)

def m_elbo_naive(model, x):
    """Computes E_{p(x)}[ELBO] for multi-modal vae --- NOT EXPOSED"""
    qz_xs, px_zs, zss = model(x)
    lpx_zs, klds = [], []
    for r, qz_x in enumerate(qz_xs):
        kld = kl_divergence(qz_x, model.pz(*model._get_pz_params))
        klds.append(kld.sum(-1))
        for d, px_z in enumerate(px_zs[r]):
            lpx_z = px_z.log_prob(x[d]) * model.vaes[d].llik_scaling
            lpx_zs.append(lpx_z.sum(-1))
    obj = (1 / len(model.vaes)) * (torch.stack(lpx_zs).sum(0) - torch.stack(klds).sum(0))
    return obj.sum()

def m_elbo_naive_warmup(model, x, beta):
    """Computes E_{p(x)}[ELBO] for multi-modal vae --- NOT EXPOSED"""
    qz_xs, px_zs, zss = model(x)
    lpx_zs, klds = [], []
    for r, qz_x in enumerate(qz_xs):
        kld = kl_divergence(qz_x, model.pz(*model._get_pz_params))
        klds.append(kld.sum(-1))
        for d, px_z in enumerate(px_zs[r]):
            lpx_z = px_z.log_prob(x[d]) * model.vaes[d].llik_scaling
            lpx_zs.append(lpx_z.sum(-1))
    obj = (1 / len(model.vaes)) * (torch.stack(lpx_zs).sum(0) - beta * torch.stack(klds).sum(0))
    return obj.sum()

def protein_preprocessing(t1):
    t0 = t1.clone()
    t0[t0 == 0] = 1
    return torch.log1p(t1 / torch.exp(torch.sum(torch.log(t0), axis=1) * (1 / torch.sum(t1 > 0, axis=1))).unsqueeze(-1))

def atac_preprocessing(t1):
    t1[t1>0] = 1
    return t1

# TODO: Not implemented
def rna_preprocessing(t1):
    return t1

class Constants(object):
    eta = 1e-6
    eps = 1e-7
    log2 = math.log(2)
    log2pi = math.log(2 * math.pi)
    logceilc = 88  # largest cuda v s.t. exp(v) < inf
    logfloorc = -104  # smallest cuda v s.t. exp(v) > 0

class ZINB(ZeroInflatedNegativeBinomial):
    def __init__(self, total_count, probs, gate):
        super(ZINB, self).__init__(total_count=total_count, probs=probs, gate=gate)

class VAE(nn.Module):
    def __init__(self, prior_dist, likelihood_dist, post_dist, enc, dec, params):
        super(VAE, self).__init__()
        self.pz = prior_dist
        self.px_z = likelihood_dist
        self.qz_x = post_dist
        self.enc = enc
        self.dec = dec
        self.modelName = None
        self.params = params
        self._pz_params = None  # defined in subclass
        self._qz_x_params = None  # populated in `forward`
        self.llik_scaling = 1.0

    @property
    def pz_params(self):
        return self._pz_params

    @property
    def qz_x_params(self):
        if self._qz_x_params is None:
            raise NameError("qz_x params not initalised yet!")
        return self._qz_x_params

    @staticmethod
    def getDataLoaders(batch_size, shuffle=True, device="cuda"):
        # handle merging individual datasets appropriately in sub-class
        raise NotImplementedError

    def forward(self, x):
        self._qz_x_params = self.enc(x)
        qz_x = self.qz_x(*self._qz_x_params)
        zs = qz_x.rsample()
        px_z = self.px_z(*self.dec(zs))
        return qz_x, px_z, zs

    def reconstruct(self, data):
        self.eval()
        with torch.no_grad():
            qz_x = self.qz_x(*self.enc(data))
            latents = qz_x.rsample()  # no dim expansion
            px_z = self.px_z(*self.dec(latents))
            recon = get_mean(px_z)
        return recon

    def reconstruct_sample(self, data):
        self.eval()
        with torch.no_grad():
            qz_x = self.qz_x(*self.enc(data))
            latents = qz_x.rsample()  # no dim expansion
            px_z = self.px_z(*self.dec(latents))
            recon = px_z._sample()
        return recon

    def latents(self, data, sampling=False):
        self.eval()
        with torch.no_grad():
            qz_x = self.qz_x(*self.enc(data))
            if not sampling:
                lats = get_mean(qz_x)
            else:
                lats = qz_x._sample()
        return lats

class Enc(nn.Module):

    def __init__(self, data_dim, latent_dim, num_hidden_layers, hidden_dim):  # added hidden_dim
        super(Enc, self).__init__()
        self.data_dim = data_dim
        modules = []
        modules.append(nn.Sequential(nn.Linear(data_dim, hidden_dim), nn.BatchNorm1d(hidden_dim), nn.ReLU(True)))
        for _ in range(num_hidden_layers - 1):
            modules.append(nn.Sequential(nn.Linear(hidden_dim, hidden_dim), nn.BatchNorm1d(hidden_dim), nn.ReLU(True)))
        self.enc = nn.Sequential(*modules)
        self.fc21 = nn.Linear(hidden_dim, latent_dim)
        self.fc22 = nn.Linear(hidden_dim, latent_dim)
        self.scale_factor = 10000

    def read_count(self, x):
        read = torch.sum(x, axis=1)
        read = read.repeat(self.data_dim, 1).t()
        return (read)

    def forward(self, x):
        read = self.read_count(x)
        x = x / read * self.scale_factor
        e = self.enc(x)
        lv = self.fc22(e).clamp(-12, 12)  # restrict to avoid torch.exp() over/underflow
        return self.fc21(e), F.softmax(lv, dim=-1) * lv.size(-1) + Constants.eta

class Dec(nn.Module):
    """ Generate an MNIST image given a sample from the latent space. """

    def __init__(self, data_dim, latent_dim, num_hidden_layers, hidden_dim, modality):  # added hidden_dim
        super(Dec, self).__init__()
        self.modality = modality
        self.data_dim = data_dim

        modules = []
        modules.append(nn.Sequential(nn.Linear(latent_dim, hidden_dim), nn.BatchNorm1d(hidden_dim), nn.ReLU(True)))
        for _ in range(num_hidden_layers - 1):
            modules.append(nn.Sequential(nn.Linear(hidden_dim, hidden_dim), nn.BatchNorm1d(hidden_dim), nn.ReLU(True)))
        self.dec = nn.Sequential(*modules)
        self.fc31 = nn.Linear(hidden_dim, data_dim)
        self.fc32 = nn.Linear(hidden_dim, data_dim)

        if self.modality == 'atac':
            # zero-inflated
            self.fc33 = nn.Linear(hidden_dim, data_dim)

    def forward(self, z):
        d = self.dec(z)
        log_r = self.fc31(d).clamp(-12, 12)  # restrict to avoid torch.exp() over/underflow
        r = torch.exp(log_r)
        p = self.fc32(d)
        p = torch.sigmoid(p).clamp(Constants.eps, 1 - Constants.eps)  # restrict to avoid probs = 0,1

        if self.modality == 'atac':
            g = self.fc33(d)
            g = torch.sigmoid(g)
            return r, p, g
        else:
            return r, p

class ATAC(VAE):
    """ Derive a specific sub-class of a VAE for ATAC. """

    def __init__(self, params):
        super(ATAC, self).__init__(
            dist.Laplace,
            ZINB,
            dist.Laplace,
            Enc(params.p_dim, params.latent_dim, params.num_hidden_layers, params.p_hidden_dim),
            Dec(params.p_dim, params.latent_dim, params.num_hidden_layers, params.p_hidden_dim, 'atac'),
            params
        )
        grad = {'requires_grad': params.learn_prior}
        self._pz_params = nn.ParameterList([
            nn.Parameter(torch.zeros(1, params.latent_dim), requires_grad=False),  # mu
            nn.Parameter(torch.zeros(1, params.latent_dim), **grad)  # logvar
        ])
        self.modelName = 'atac'
        self.data_dim = self.params.p_dim
        self.llik_scaling = 1.
        self.scale_factor = 10000

    @property
    def pz_params(self):
        return self._pz_params[0], F.softmax(self._pz_params[1], dim=1) * self._pz_params[1].size(-1)

    @staticmethod
    def getDataLoaders(dataset, batch_size, shuffle=True, device="cuda"):
        kwargs = {'num_workers': 1, 'pin_memory': True} if device == "cuda" else {}
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, drop_last=False, **kwargs)
        return dataloader

    def forward(self, x):
        read_count = self.enc.read_count(x)
        self._qz_x_params = self.enc(x)
        qz_x = self.qz_x(*self._qz_x_params)
        zs = qz_x.rsample()
        r, p, g = self.dec(zs)
        r = r / self.scale_factor * read_count
        px_z = self.px_z(r, p, g)
        return qz_x, px_z, zs

class Protein(VAE):
    """ Derive a specific sub-class of a VAE for Protein. """
    def __init__(self, params):
        super(Protein, self).__init__(
            dist.Laplace,
            dist.NegativeBinomial,
            dist.Laplace,
            Enc(params.p_dim, params.latent_dim, params.num_hidden_layers, params.p_hidden_dim),
            Dec(params.p_dim, params.latent_dim, params.num_hidden_layers, params.p_hidden_dim, 'protein'),
            params
        )
        grad = {'requires_grad': params.learn_prior}
        self._pz_params = nn.ParameterList([
            nn.Parameter(torch.zeros(1, params.latent_dim), requires_grad=False),  # mu
            nn.Parameter(torch.zeros(1, params.latent_dim), **grad)  # logvar
        ])
        self.modelName = 'protein'
        self.data_dim = self.params.p_dim
        self.llik_scaling = 1.
        self.scale_factor = 10000

    @property
    def pz_params(self):
        return self._pz_params[0], F.softmax(self._pz_params[1], dim=1) * self._pz_params[1].size(-1)

    @staticmethod
    def getDataLoaders(dataset, batch_size, shuffle=True, device="cuda"):
        kwargs = {'num_workers': 1, 'pin_memory': True} if device == "cuda" else {}
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, drop_last=False, **kwargs)

        return dataloader

    def forward(self, x):
        read_count = self.enc.read_count(x)
        self._qz_x_params = self.enc(x)
        qz_x = self.qz_x(*self._qz_x_params)
        zs = qz_x.rsample()
        r, _ = self.dec(zs)
        r = r / self.scale_factor * read_count
        px_z = self.px_z(r, _)
        return qz_x, px_z, zs

class RNA(VAE):
    """ Derive a specific sub-class of a VAE for RNA. """

    def __init__(self, params):
        super(RNA, self).__init__(
            dist.Laplace,
            dist.NegativeBinomial,  # likelihood
            dist.Laplace,
            Enc(params.r_dim, params.latent_dim, params.num_hidden_layers, params.r_hidden_dim),
            Dec(params.r_dim, params.latent_dim, params.num_hidden_layers, params.r_hidden_dim, 'rna'),
            params
        )
        grad = {'requires_grad': params.learn_prior}
        self._pz_params = nn.ParameterList([
            nn.Parameter(torch.zeros(1, params.latent_dim), requires_grad=False),  # mu
            nn.Parameter(torch.zeros(1, params.latent_dim), **grad)  # logvar
        ])
        self.modelName = 'rna'
        self.data_dim = self.params.r_dim
        self.llik_scaling = 1.
        self.scale_factor = 10000

    @property
    def pz_params(self):
        return self._pz_params[0], F.softmax(self._pz_params[1], dim=1) * self._pz_params[1].size(-1)

    @staticmethod
    def getDataLoaders(dataset, batch_size, shuffle=True, device="cuda"):
        kwargs = {'num_workers': 1, 'pin_memory': True} if device == "cuda" else {}
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, drop_last=False, **kwargs)

        return dataloader

    def forward(self, x):
        read_count = self.enc.read_count(x)
        self._qz_x_params = self.enc(x)
        qz_x = self.qz_x(*self._qz_x_params)
        zs = qz_x.rsample()
        r, _ = self.dec(zs)
        r = r / self.scale_factor * read_count
        px_z = self.px_z(r, _)
        return qz_x, px_z, zs

class MMVAE(nn.Module):
    """MMVAE class.

    Parameters
    ----------
    subtask : str
        Name of the subtask which is composed of the name of two modality. This parameter will indicate some modality-specific features in the model.
    params : argparse.Namespace
        A Namespace object that contains arguments of MMVAE. For details of parameters in parser args, please refer to link (parser help document).
    """

    def __init__(self, subtask, params):
        super(MMVAE, self).__init__()
        self.pz = dist.Laplace
        assert subtask in ('rna-dna', 'rna-protein')
        self.modelName = subtask
        if subtask == 'rna-dna':
            self.preprocessing = atac_preprocessing
            self.vaes = nn.ModuleList([RNA(params), ATAC(params)])
        else:
            self.preprocessing = protein_preprocessing
            self.vaes = nn.ModuleList([RNA(params), Protein(params)])
        self.params = params

        grad = {'requires_grad': params.learn_prior}
        self._pz_params = nn.ParameterList([
            nn.Parameter(torch.zeros(1, params.latent_dim), requires_grad=False),  # mu
            nn.Parameter(torch.zeros(1, params.latent_dim), **grad)  # logvar
        ])
        self.vaes[0].llik_scaling = prod(self.vaes[1].dataSize) / prod(self.vaes[0].dataSize) \
            if params.llik_scaling == 0 else params.llik_scaling
        self.scale_factor = 10000

    @property
    def _get_pz_params(self):
        return self._pz_params[0], F.softmax(self._pz_params[1], dim=1) * self._pz_params[1].size(-1)

    def _get_cluster(self, data, modality='both', n_clusters=10, method='kmeans', device='cuda'):
        self.eval()
        lats = self._get_latents(data, sampling=False)
        if modality == 'both':
            lat = sum(lats) / len(lats)
        elif modality == 'rna':
            lat = lats[0]
        elif modality == 'atac':
            lat = lats[1]

        if method == 'kmeans':
            fit = KMeans(n_clusters=n_clusters, random_state=0, init='k-means++').fit(lat.cpu().numpy())
            cluster = fit.labels_
        elif method == 'dbscan':
            fit = DBSCAN(eps=0.5, min_samples=50).fit(lat.cpu().numpy())
            cluster = fit.labels_
        else:
            gamma, _, _, _, _ = self.get_gamma(lat)
            cluster = torch.argmax(gamma, axis=1)
            cluster = cluster.detach().numpy()
            fit = None

        return cluster, fit

    def forward(self, x):
        """Forward function for torch.nn.Module.

        Parameters
        ----------
        x : list[torch.Tensor]
            Features of two modalities.

        Returns
        -------
        qz_xs : list[torch.Tensor]
            Post prior of two modalities.
        px_zs : list[torch.Tensor]
            likelihood of two modalities.
        zss : list[torch.Tensor]
            Reconstruction results of two modalities.
        """

        qz_xs, zss = [], []
        read_counts = []
        # initialise cross-modal matrix
        px_zs = [[None for _ in range(len(self.vaes))] for _ in range(len(self.vaes))]
        for m, vae in enumerate(self.vaes):
            read_counts.append(vae.enc.read_count(x[m]))
            qz_x, px_z, zs = vae(x[m])
            qz_xs.append(qz_x)
            zss.append(zs)
            px_zs[m][m] = px_z  # fill-in diagonal
        for e, zs in enumerate(zss):
            for d, vae in enumerate(self.vaes):
                if e != d:  # fill-in off-diagonal
                    if self.modelName == 'rna-protein':
                        r, _ = vae.dec(zs)
                        r = r / self.scale_factor * read_counts[d]
                        px_zs[e][d] = vae.px_z(r, _)
                    else:
                        if d == 0:
                            r, p = vae.dec(zs)
                            r = r / self.scale_factor * read_counts[d]
                            px_zs[e][d] = vae.px_z(r, p)
                        else:
                            r, p, g = vae.dec(zs)
                            r = r / self.scale_factor * read_counts[d]
                            px_zs[e][d] = vae.px_z(r, p, g)

        return qz_xs, px_zs, zss

    def _reconstruct(self, data):
        self.eval()
        with torch.no_grad():
            _, px_zs, _ = self.forward(data)
            # cross-modal matrix of reconstructions of MEANS
            recons = [[get_mean(px_z) for px_z in r] for r in px_zs]
        return recons

    def _reconstruct_sample(self, data):
        self.eval()
        with torch.no_grad():
            _, px_zs, _ = self.forward(data)
            # cross-modal matrix of reconstructions of SAMPLES
            recons = [[px_z.sample() for px_z in r] for r in px_zs]
        return recons

    def _get_latents(self, data, sampling=False):
        self.eval()
        with torch.no_grad():
            qz_xs, _, _ = self.forward(data)
            if not sampling:
                lats = [get_mean(qz_x) for qz_x in qz_xs]
            else:
                lats = [qz_x._sample() for qz_x in qz_xs]
        return lats

    def fit(self, dataset):
        """fit function for training.

        Parameters
        ----------
        dataset : dance.datasets.multimodality.ModalityMatchingDataset
            Dataset for modality matching.

        Returns
        -------
        None.
        """

        start_early_stop = self.params.deterministic_warmup

        idx = np.random.permutation(dataset.modalities[0].shape[0])
        train_idx = idx[:int(idx.shape[0] * 0.85)]
        val_idx = idx[int(idx.shape[0] * 0.85):]

        optimizer = optim.Adam(filter(lambda p: p.requires_grad, self.parameters()),
                               lr=self.params.lr, amsgrad=True)
        assert (self.params.obj in ['m_elbo_naive', 'm_elbo_naive_warmup'])
        objective = m_elbo_naive_warmup if self.params.obj == 'm_elbo_naive_warmup' else m_elbo_naive
        train_dataset = SimpleIndexDataset(train_idx)
        train_loader = DataLoader(
            dataset=train_dataset,
            batch_size=self.params.batch_size,
            shuffle=True,
            num_workers=0,
            drop_last=True,
        )

        train_mod1 = torch.from_numpy(dataset.numpy_features(0, True)).float().to(self.params.device)
        train_mod2 = torch.from_numpy(dataset.numpy_features(1, True)).float().to(self.params.device)
        test_mod1 = torch.from_numpy(dataset.numpy_features(2, True)).float().to(self.params.device)
        test_mod2 = torch.from_numpy(dataset.numpy_features(3, True)).float().to(self.params.device)
        labels = torch.from_numpy(dataset.test_sol.X.toarray())
        vals = []
        tr = []

        for epoch in range(1, self.params.epochs + 1):
            self.train()
            b_loss = 0
            for i, batch_idx in enumerate(train_loader):
                dataT = (train_mod1[batch_idx], train_mod2[batch_idx])
                beta = (epoch - 1) / start_early_stop if epoch <= start_early_stop else 1
                if dataT[0].size()[0] == 1:
                    continue
                # data = [d.to(self.paradevice) for d in dataT]  # multimodal
                data = dataT
                optimizer.zero_grad()
                loss = -objective(self, data, beta)
                loss.backward()
                optimizer.step()
                b_loss += loss.item()
                if self.params.print_freq > 0 and i % self.params.print_freq == 0:
                    print("iteration {:04d}: loss: {:6.3f}".format(i, loss.item() / self.params.batch_size))
            tr.append(b_loss / len(train_loader.dataset))
            print('====> Epoch: {:03d} Train loss: {:.4f}'.format(epoch, tr[-1]))

            if torch.isnan(torch.tensor([b_loss])):
                break

            vals.append(self.score(train_mod1[val_idx], train_mod2[val_idx], metric='loss'))
            print('====>             Valid loss: {:.4f}'.format(vals[-1]))

            if vals[-1] == min(vals):
                if not os.path.exists('models'):
                    os.mkdir('models')
                torch.save(self.state_dict(), f'models/model_{self.params.rnd_seed}.pth')

            if epoch % 10 == 0:
                print('Valid Matching score:', self.score(train_mod1[val_idx], train_mod2[val_idx], torch.eye(val_idx.shape[0])))
                print('Test Matching score:', self.score(test_mod1, test_mod2, labels))

            if epoch > start_early_stop and min(vals) != min(vals[-10:]):
                print('Early stopped.')
                self.load_state_dict(torch.load(f'models/model_{self.params.rnd_seed}.pth'))
                break

    def score(self, mod1, mod2, labels=None, metric='minkowski'):
        """Score function to get score of prediction.

        Parameters
        ----------
        mod1 : torch.Tensor
            Features of modality 1.
        mod2 : torch.Tensor
            Features of modality 2.
        labels : torch.Tensor optional
            Labels of matching modality, i.e. cell correspondence between two modalities. Required when metric is not 'loss'.
        metric : str optional
            Metric of the score function, by default to be 'minkowski'.

        Returns
        -------
        score : float
            Score of predicted matching, according to specified metric.

        """
        self.eval()
        if labels is None:
            assert metric=='loss', 'Unable to evaluate without labels.'
        if metric == 'loss':
            b_loss = 0
            idx = np.arange(mod1.shape[0])
            dataset = SimpleIndexDataset(idx)
            data_loader = DataLoader(
                dataset=dataset,
                batch_size=self.params.batch_size,
                shuffle=False,
                num_workers=0,
                drop_last=False,
            )
            with torch.no_grad():
                for i, batch_idx in enumerate(data_loader):
                    objective = m_elbo_naive_warmup if self.params.obj == 'm_elbo_naive_warmup' else 'm_elbo_naive'
                    loss = -objective(self, [mod1[batch_idx], mod2[batch_idx]], 1).item()
                    b_loss += loss
            return b_loss / mod1.shape[0]
        else:
            pred = self.predict([mod1, mod2], metric=metric)
        return ((pred * labels).sum() / mod1.shape[0]).item()


    def predict(self, feats, metric ='minkowski'):
        """Predict function to get score of prediction.

        Parameters
        ----------
        feats : list[torch.Tensor]
            Features of two modalities.
        metric : str optional
            Metric of the matching function, by default to be 'minkowski'.

        Returns
        -------
        pred : float
            Predicted matching between two modalities.

        """
        self.eval()
        idx = np.arange(feats[0].shape[0])
        dataset = SimpleIndexDataset(idx)
        data_loader = DataLoader(
            dataset=dataset,
            batch_size=self.params.batch_size*10,
            shuffle=False,
            num_workers=0,
            drop_last=False,
        )
        pred = []
        with torch.no_grad():
            for i, batch_idx in enumerate(data_loader):
                dataT = [feats[0][batch_idx], feats[1][batch_idx]]
                lats = self._get_latents(dataT, sampling=False)
                if i == 0:
                    pred = lats
                else:
                    for m, lat in enumerate(lats):
                        pred[m] = torch.cat([pred[m], lat], dim=0)

            lats = []
            for m, lat in enumerate(pred):
                lat = lat.cpu()
                lats.append(lat)

        nn = NearestNeighbors(metric=metric)
        nn.fit(lats[0])
        transp_nearest_neighbor = torch.tensor(nn.kneighbors(lats[1], 1, return_distance=False))
        pred = torch.zeros(lats[0].shape[0], lats[0].shape[0])
        pred[torch.arange(lats[0].shape[0]), transp_nearest_neighbor.squeeze(-1)] = 1
        return pred




import os
from dance.data import *
import anndata as ad
import scanpy as sc
import pickle
import numpy as np
import torch
from dance.transforms.preprocess import lsiTransformer

class MultiModalityDataset():
    def __init__(self, task, data_url, subtask, data_dir = "./data"):

        assert (subtask in ['openproblems_bmmc_multiome_phase2_mod2', 'openproblems_bmmc_multiome_phase2_rna',
                            'openproblems_bmmc_cite_phase2_rna', 'openproblems_bmmc_cite_phase2_mod2',
                            'openproblems_bmmc_cite_phase2', 'openproblems_bmmc_multiome_phase2',
                            'adt2gex', 'gex2adt',
                            'atac2gex', 'gex2atac']), 'Undefined subtask.'

        assert (task in ['predict_modality', 'match_modality', 'joint_embedding']), 'Undefined task.'

        # regularize subtask name
        if task == 'joint_embedding':
            if subtask == 'adt':
                subtask = 'openproblems_bmmc_cite_phase2'
            elif subtask == 'atac':
                subtask = 'openproblems_bmmc_multiome_phase2'
        else:
            if subtask == 'adt2gex':
                subtask = 'openproblems_bmmc_cite_phase2_mod2'
            elif subtask == 'gex2adt':
                subtask = 'openproblems_bmmc_cite_phase2_rna'
            elif subtask == 'atac2gex':
                subtask = 'openproblems_bmmc_multiome_phase2_mod2'
            elif subtask == 'gex2atac':
                subtask = 'openproblems_bmmc_multiome_phase2_rna'

        self.task = task
        self.subtask = subtask
        self.data_dir = data_dir
        self.loaded = False
        self.data_url = data_url

    def download_data(self):
        # download data
        download_file(self.data_url, self.data_dir+"/{}.zip".format(self.subtask))
        unzip_file(self.data_dir+"/{}.zip".format(self.subtask), self.data_dir)
        return self

    def download_pathway(self):
        download_file('https://www.dropbox.com/s/uqoakpalr3albiq/h.all.v7.4.entrez.gmt?dl=1', self.data_dir+"/h.all.v7.4.entrez.gmt")
        download_file('https://www.dropbox.com/s/yjrcsd2rpmahmfo/h.all.v7.4.symbols.gmt?dl=1', self.data_dir+"/h.all.v7.4.symbols.gmt")
        return self

    def is_complete(self):
        # judge data is complete or not
        if self.task == 'joint_embedding':
            return os.path.exists(
                os.path.join(self.data_dir, self.subtask, f'{self.subtask}.censor_dataset.output_mod1.h5ad')
            ) and os.path.exists(
                os.path.join(self.data_dir, self.subtask, f'{self.subtask}.censor_dataset.output_mod2.h5ad'))
        else:
            return os.path.exists(
                os.path.join(self.data_dir, self.subtask, f'{self.subtask}.censor_dataset.output_train_mod1.h5ad')
            ) and os.path.exists(
                os.path.join(self.data_dir, self.subtask, f'{self.subtask}.censor_dataset.output_train_mod2.h5ad')
            ) and os.path.exists(
                os.path.join(self.data_dir, self.subtask, f'{self.subtask}.censor_dataset.output_test_mod1.h5ad')
            ) and os.path.exists(
                os.path.join(self.data_dir, self.subtask, f'{self.subtask}.censor_dataset.output_test_mod2.h5ad'))


    def load_data(self):
        # Load data from existing h5ad files, or download files and load data.
        if self.is_complete():
            pass
        else:
            self.download_data()
            assert self.is_complete()

        if self.task == 'joint_embedding':
            mod_path_list = [os.path.join(self.data_dir, self.subtask, f'{self.subtask}.censor_dataset.output_mod1.h5ad'),
                os.path.join(self.data_dir, self.subtask, f'{self.subtask}.censor_dataset.output_mod2.h5ad')]
        else:
            mod_path_list = [os.path.join(self.data_dir, self.subtask, f'{self.subtask}.censor_dataset.output_train_mod1.h5ad'),
                os.path.join(self.data_dir, self.subtask, f'{self.subtask}.censor_dataset.output_train_mod2.h5ad'),
                os.path.join(self.data_dir, self.subtask, f'{self.subtask}.censor_dataset.output_test_mod1.h5ad'),
                os.path.join(self.data_dir, self.subtask, f'{self.subtask}.censor_dataset.output_test_mod2.h5ad')]

        self.modalities = []
        for mod_path in mod_path_list:
            self.modalities.append(ad.read_h5ad(mod_path))
        self.loaded = True
        return self

    def sparse_features(self, index=None, count=False):
        assert self.loaded, 'Data have not been loaded.'
        if not count:
            if index is None:
                return [mod.X for mod in self.modalities]
            else:
                return self.modalities[index].X
        else:
            if index is None:
                return [mod.layers['counts'] for mod in self.modalities]
            else:
                return self.modalities[index].layers['counts']

    def numpy_features(self, index=None, count = False):
        assert self.loaded, 'Data have not been loaded.'
        if not count:
            if index is None:
                return [mod.X.toarray() for mod in self.modalities]
            else:
                return self.modalities[index].X.toarray()
        else:
            if index is None:
                return [mod.layers['counts'].toarray() for mod in self.modalities]
            else:
                return self.modalities[index].layers['counts'].toarray()

    def tensor_features(self, index=None, count = False, device='cpu'):
        assert self.loaded, 'Data have not been loaded.'
        if not count:
            if index is None:
                return [torch.from_numpy(mod.X.toarray()).to(device) for mod in self.modalities]
            else:
                return torch.from_numpy(self.modalities[index].X.toarray()).to(device)
        else:
            if index is None:
                return [torch.from_numpy(mod.layers['counts'].toarray()).to(device) for mod in self.modalities]
            else:
                return torch.from_numpy(self.modalities[index].layers['counts'].toarray()).to(device)

    def get_modalities(self):
        assert self.loaded, 'Data have not been loaded.'
        return self.modalities

class ModalityPredictionDataset(MultiModalityDataset):
    def __init__(self, subtask, data_dir = "./data"):
        assert (subtask in ['openproblems_bmmc_multiome_phase2_mod2', 'openproblems_bmmc_multiome_phase2_rna',
                             'openproblems_bmmc_cite_phase2_rna', 'openproblems_bmmc_cite_phase2_mod2',
                             'adt2gex', 'gex2adt',
                             'atac2gex', 'gex2atac']), 'Undefined subtask.'

        if subtask == 'adt2gex':
            subtask = 'openproblems_bmmc_cite_phase2_mod2'
        elif subtask == 'gex2adt':
            subtask = 'openproblems_bmmc_cite_phase2_rna'
        elif subtask == 'atac2gex':
            subtask = 'openproblems_bmmc_multiome_phase2_mod2'
        elif subtask == 'gex2atac':
            subtask = 'openproblems_bmmc_multiome_phase2_rna'

        data_url = {'openproblems_bmmc_cite_phase2_mod2': 'https://www.dropbox.com/s/snh8knscnlcq4um/openproblems_bmmc_cite_phase2_mod2.zip?dl=1',
                    'openproblems_bmmc_cite_phase2_rna': 'https://www.dropbox.com/s/xbfyhv830u9pupv/openproblems_bmmc_cite_phase2_rna.zip?dl=1',
                    'openproblems_bmmc_multiome_phase2_mod2': 'https://www.dropbox.com/s/p9ve2ljyy4yqna4/openproblems_bmmc_multiome_phase2_mod2.zip?dl=1',
                    'openproblems_bmmc_multiome_phase2_rna': 'https://www.dropbox.com/s/cz60vp7bwapz0kw/openproblems_bmmc_multiome_phase2_rna.zip?dl=1'}

        super(ModalityPredictionDataset, self).__init__('predict_modality', data_url.get(subtask), subtask, data_dir)

    def preprocess(self, kind='feature_selection', selection_threshold=10000):
        if kind == 'pca':
            print('Preprocessing method not supported.')
            return self
        elif kind =='feature_selection':
            if self.modalities[0].shape[1]>selection_threshold:
                sc.pp.highly_variable_genes(self.modalities[0], layer='counts', flavor='seurat_v3', n_top_genes=selection_threshold)
                self.modalities[2].var['highly_variable'] = self.modalities[0].var['highly_variable']
                for i in [0, 2]:
                    self.modalities[i] = self.modalities[i][:, self.modalities[i].var['highly_variable']]
        else:
            print('Preprocessing method not supported.')
            return self
        print('Preprocessing done.')
        return self


class ModalityMatchingDataset(MultiModalityDataset):
    def __init__(self, subtask, data_dir = "./data"):
        assert (subtask in ['openproblems_bmmc_multiome_phase2_mod2', 'openproblems_bmmc_multiome_phase2_rna',
                             'openproblems_bmmc_cite_phase2_rna', 'openproblems_bmmc_cite_phase2_mod2',
                             'adt2gex', 'gex2adt',
                             'atac2gex', 'gex2atac']), 'Undefined subtask.'

        if subtask == 'adt2gex':
            subtask = 'openproblems_bmmc_cite_phase2_mod2'
        elif subtask == 'gex2adt':
            subtask = 'openproblems_bmmc_cite_phase2_rna'
        elif subtask == 'atac2gex':
            subtask = 'openproblems_bmmc_multiome_phase2_mod2'
        elif subtask == 'gex2atac':
            subtask = 'openproblems_bmmc_multiome_phase2_rna'

        data_url = {'openproblems_bmmc_cite_phase2_mod2': 'https://www.dropbox.com/s/fa6zut89xx73itz/openproblems_bmmc_cite_phase2_mod2.zip?dl=1',
                    'openproblems_bmmc_cite_phase2_rna': 'https://www.dropbox.com/s/ep00mqcjmdu0b7v/openproblems_bmmc_cite_phase2_rna.zip?dl=1',
                    'openproblems_bmmc_multiome_phase2_mod2': 'https://www.dropbox.com/s/31qi5sckx768acw/openproblems_bmmc_multiome_phase2_mod2.zip?dl=1',
                    'openproblems_bmmc_multiome_phase2_rna': 'https://www.dropbox.com/s/h1s067wkefs1jh2/openproblems_bmmc_multiome_phase2_rna.zip?dl=1'}

        super(ModalityMatchingDataset, self).__init__('match_modality', data_url.get(subtask), subtask, data_dir)
        self.preprocessed = False

    def load_sol(self):
        assert(self.loaded)
        self.train_sol = ad.read_h5ad(os.path.join(self.data_dir, self.subtask, f'{self.subtask}.censor_dataset.output_train_sol.h5ad'))
        self.test_sol = ad.read_h5ad(os.path.join(self.data_dir, self.subtask, f'{self.subtask}.censor_dataset.output_test_sol.h5ad'))
        self.modalities[1] = self.modalities[1][self.train_sol.to_df().values.argmax(1)]
        return self

    def preprocess(self, kind = 'pca', pkl_path=None, selection_threshold=10000):

        # TODO: support other two subtasks
        assert self.subtask in ('openproblems_bmmc_cite_phase2_rna', 'openproblems_bmmc_multiome_phase2_rna'), 'Currently not available.'

        if kind == 'pca':
            if pkl_path and (not os.path.exists(pkl_path)):

                if self.subtask == 'openproblems_bmmc_cite_phase2_rna':
                    lsi_transformer_gex = lsiTransformer(n_components=256, drop_first=True)
                    m1_train = lsi_transformer_gex.fit_transform(self.modalities[0]).values
                    m1_test = lsi_transformer_gex.transform(self.modalities[2]).values
                    m2_train = self.modalities[1].X.toarray()
                    m2_test = self.modalities[3].X.toarray()

                if self.subtask == 'openproblems_bmmc_multiome_phase2_rna':
                    lsi_transformer_gex = lsiTransformer(n_components=256, drop_first=True)
                    m1_train = lsi_transformer_gex.fit_transform(self.modalities[0]).values
                    m1_test = lsi_transformer_gex.transform(self.modalities[2]).values
                    lsi_transformer_atac = lsiTransformer(n_components=512, drop_first=True)
                    m2_train = lsi_transformer_atac.fit_transform(self.modalities[1]).values
                    m2_test = lsi_transformer_atac.transform(self.modalities[3]).values

                self.preprocessed_features = {'mod1_train': m1_train, 'mod2_train': m2_train,
                                              'mod1_test': m1_test, 'mod2_test': m2_test}
                pickle.dump(self.preprocessed_features, open(pkl_path, 'wb'))

            else:
                self.preprocessed_features = pickle.load(open(pkl_path, 'rb'))
        elif kind == 'feature_selection':
            for i in range(2):
                if self.modalities[i].shape[1] > selection_threshold:
                    sc.pp.highly_variable_genes(self.modalities[i], layer='counts', flavor='seurat_v3', n_top_genes=selection_threshold)
                    self.modalities[i+2].var['highly_variable'] = self.modalities[i].var['highly_variable']
                    self.modalities[i] = self.modalities[i][:, self.modalities[i].var['highly_variable']]
                    self.modalities[i+2] = self.modalities[i+2][:, self.modalities[i+2].var['highly_variable']]
        else:
            print('Preprocessing method not supported.')
            return self
        print('Preprocessing done.')
        self.preprocessed = True
        return self

    def get_preprocessed_features(self):
        assert self.preprocessed, 'Transformed features do not exist.'
        return self.preprocessed_features


class JointEmbeddingNIPSDataset(MultiModalityDataset):
    def __init__(self, subtask, data_dir = "./data"):
        assert (subtask in ['openproblems_bmmc_multiome_phase2',
                             'openproblems_bmmc_cite_phase2',
                             'adt', 'atac']), 'Undefined subtask.'

        if subtask == 'adt':
            subtask = 'openproblems_bmmc_cite_phase2'
        elif subtask == 'atac':
            subtask = 'openproblems_bmmc_multiome_phase2'

        data_url = {'openproblems_bmmc_cite_phase2': 'https://www.dropbox.com/s/hjr4dxuw55vin5z/openproblems_bmmc_cite_phase2.zip?dl=1',
                    'openproblems_bmmc_multiome_phase2': 'https://www.dropbox.com/s/40kjslupxhkg92s/openproblems_bmmc_multiome_phase2.zip?dl=1'}
        super(JointEmbeddingNIPSDataset, self).__init__('joint_embedding', data_url.get(subtask), subtask, data_dir)
        self.preprocessed = False

    def load_metadata(self):
        assert(self.loaded)

        if self.subtask.find('cite')!=-1:
            mod = 'adt'
            meta = 'cite'
        else:
            mod = 'atac'
            meta = 'multiome'
        self.exploration = [ad.read_h5ad(os.path.join(self.data_dir, self.subtask, f'{meta}_gex_processed_training.h5ad')),
                            ad.read_h5ad(os.path.join(self.data_dir, self.subtask, f'{meta}_{mod}_processed_training.h5ad')),]
        return self

    def load_sol(self):
        assert(self.loaded)
        self.test_sol = ad.read_h5ad(os.path.join(self.data_dir, self.subtask, f'{self.subtask}.censor_dataset.output_solution.h5ad'))
        return self

    def preprocess(self, kind='aux', pretrained_folder = '.', selection_threshold=10000):
        if kind == 'aux':
            os.makedirs(pretrained_folder, exist_ok=True)

            if os.path.exists(os.path.join(pretrained_folder, f'preprocessed_data_{self.subtask}.pkl')):

                with open(os.path.join(pretrained_folder, f'preprocessed_data_{self.subtask}.pkl'), 'rb') as f:
                    self.preprocessed_data = pickle.load(f)
                with open(os.path.join(pretrained_folder, f'{self.subtask}_config.pk'), 'rb') as f:
                    self.nb_cell_types, self.nb_batches, self.nb_phases = pickle.load(f)
                self.preprocessed = True
                print('Preprocessing done.')
                return self

            ##########################################
            ##             PCA PRETRAIN             ##
            ##########################################

            # scale and log transform
            scale = 1e4
            n_components_mod1, n_components_mod2 = 256, 100

            mod1 = self.modalities[0].var["feature_types"][0]
            mod2 = self.modalities[1].var["feature_types"][0]

            if mod2 == "ADT":
                if os.path.exists(os.path.join(pretrained_folder, f"lsi_cite_{mod1}.pkl")):
                    lsi_transformer_gex = pickle.load(open(os.path.join(pretrained_folder, f"lsi_cite_{mod1}.pkl"), "rb"))
                else:
                    lsi_transformer_gex = lsiTransformer(n_components=256, drop_first=True)
                    lsi_transformer_gex.fit(self.modalities[0])
                    pickle.dump(lsi_transformer_gex, open(os.path.join(pretrained_folder, f"lsi_cite_{mod1}.pkl"), "wb"))

            if mod2 == "ATAC":

                if os.path.exists(os.path.join(pretrained_folder, f"lsi_multiome_{mod1}.pkl")):
                    with open(os.path.join(pretrained_folder, f"lsi_multiome_{mod1}.pkl"), "rb") as f:
                        lsi_transformer_gex = pickle.load(f)
                else:
                    lsi_transformer_gex = lsiTransformer(n_components=64, drop_first=True)
                    lsi_transformer_gex.fit(self.modalities[0])
                    with open(os.path.join(pretrained_folder, f"lsi_multiome_{mod1}.pkl"), "wb") as f:
                        pickle.dump(lsi_transformer_gex, f)

                if os.path.exists(os.path.join(pretrained_folder, f"lsi_multiome_{mod2}.pkl")):
                    with open(os.path.join(pretrained_folder, f"lsi_multiome_{mod2}.pkl"), "rb") as f:
                        lsi_transformer_atac = pickle.load(f)
                else:
                    #         lsi_transformer_atac = TruncatedSVD(n_components=100, random_state=random_seed)
                    lsi_transformer_atac = lsiTransformer(n_components=512, drop_first=True)
                    lsi_transformer_atac.fit(self.modalities[1])
                    with open(os.path.join(pretrained_folder, f"lsi_multiome_{mod2}.pkl"), "wb") as f:
                        pickle.dump(lsi_transformer_atac, f)


            ##########################################
            ##           DATA PREPROCESSING         ##
            ##########################################

            ad_mod1 = self.exploration[0]
            ad_mod2 = self.exploration[1]
            mod1_obs = ad_mod1.obs

            # Make sure exploration data match the full data
            assert ((self.modalities[0].obs['batch'].index[:mod1_obs.shape[0]] == mod1_obs['batch'].index).mean() == 1)

            if mod2 == "ADT":
                mod1_pca = lsi_transformer_gex.transform(ad_mod1).values
                mod1_pca_test = lsi_transformer_gex.transform(self.modalities[0][mod1_obs.shape[0]:]).values
                mod2_pca = ad_mod2.X.toarray()
                mod2_pca_test = self.numpy_features(1)[mod1_obs.shape[0]:]

            if mod2 == "ATAC":
                mod1_pca = lsi_transformer_gex.transform(ad_mod1).values
                mod1_pca_test = lsi_transformer_gex.transform(self.modalities[0][mod1_obs.shape[0]:]).values
                mod2_pca = lsi_transformer_atac.transform(ad_mod2).values
                mod2_pca_test = lsi_transformer_atac.transform(self.modalities[1][mod1_obs.shape[0]:]).values

            cell_cycle_genes = [
                'MCM5', 'PCNA', 'TYMS', 'FEN1', 'MCM2', \
                'MCM4', 'RRM1', 'UNG', 'GINS2', 'MCM6', \
                'CDCA7', 'DTL', 'PRIM1', 'UHRF1', 'MLF1IP', \
                'HELLS', 'RFC2', 'RPA2', 'NASP', 'RAD51AP1', \
                'GMNN', 'WDR76', 'SLBP', 'CCNE2', 'UBR7', \
                'POLD3', 'MSH2', 'ATAD2', 'RAD51', 'RRM2', 'CDC45', \
                'CDC6', 'EXO1', 'TIPIN', 'DSCC1', 'BLM', 'CASP8AP2', \
                'USP1', 'CLSPN', 'POLA1', 'CHAF1B', 'BRIP1', 'E2F8', \
                'HMGB2', 'CDK1', 'NUSAP1', 'UBE2C', 'BIRC5', 'TPX2', \
                'TOP2A', 'NDC80', 'CKS2', 'NUF2', 'CKS1B', 'MKI67', \
                'TMPO', 'CENPF', 'TACC3', 'FAM64A', 'SMC4', 'CCNB2', \
                'CKAP2L', 'CKAP2', 'AURKB', 'BUB1', 'KIF11', 'ANP32E', \
                'TUBB4B', 'GTSE1', 'KIF20B', 'HJURP', 'CDCA3', 'HN1', \
                'CDC20', 'TTK', 'CDC25C', 'KIF2C', 'RANGAP1', 'NCAPD2', \
                'DLGAP5', 'CDCA2', 'CDCA8', 'ECT2', 'KIF23', 'HMMR', \
                'AURKA', 'PSRC1', 'ANLN', 'LBR', 'CKAP5', 'CENPE', 'CTCF', \
                'NEK2', 'G2E3', 'GAS2L3', 'CBX5', 'CENPA']

            print('Data loading and pca done', mod1_pca.shape, mod2_pca.shape)
            print('Start to calculate cell_cycle score. It may roughly take an hour.')

            pca_combined = np.concatenate([mod1_pca, mod2_pca], axis=1)

            del mod1_pca, mod2_pca

            cell_type_labels = mod1_obs['cell_type']
            batch_ids = mod1_obs['batch']
            phase_labels = mod1_obs['phase']
            nb_cell_types = len(np.unique(cell_type_labels))
            nb_batches = len(np.unique(batch_ids))
            nb_phases = len(np.unique(phase_labels)) - 1  # 2
            c_labels = np.array([list(np.unique(cell_type_labels)).index(item) for item in cell_type_labels])
            b_labels = np.array([list(np.unique(batch_ids)).index(item) for item in batch_ids])
            p_labels = np.array([list(np.unique(phase_labels)).index(item) for item in phase_labels])
            # 0:G1, 1:G2M, 2: S, only consider the last two
            s_genes = cell_cycle_genes[:43]
            g2m_genes = cell_cycle_genes[43:]
            sc.pp.log1p(ad_mod1)
            sc.pp.scale(ad_mod1)
            sc.tl.score_genes_cell_cycle(ad_mod1, s_genes=s_genes, g2m_genes=g2m_genes)
            S_scores = ad_mod1.obs['S_score'].values
            G2M_scores = ad_mod1.obs['G2M_score'].values
            phase_scores = np.stack([S_scores, G2M_scores]).T  # (nb_cells, 2)

            X_train = pca_combined
            # c_labels: cell type; b_labels: batch ids; p_labels: phase_labels; phase_scores
            Y_train = [c_labels, b_labels, p_labels, phase_scores]
            X_test = np.concatenate([mod1_pca_test, mod2_pca_test], axis=1)

            self.preprocessed_data = {'X_train': X_train, 'Y_train': Y_train, 'X_test': X_test}
            pickle.dump(self.preprocessed_data,
                    open(os.path.join(pretrained_folder, f'preprocessed_data_{self.subtask}.pkl'), 'wb'))
            pickle.dump([nb_cell_types, nb_batches, nb_phases],
                    open(os.path.join(pretrained_folder, f'{self.subtask}_config.pk'), 'wb'))

            self.nb_cell_types, self.nb_batches, self.nb_phases = nb_cell_types, nb_batches, nb_phases
        elif kind == 'feature_selection':
            for i in range(2):
                if self.modalities[i].shape[1] > selection_threshold:
                    sc.pp.highly_variable_genes(self.modalities[i], layer='counts', flavor='seurat_v3',
                                                n_top_genes=selection_threshold)
                    self.modalities[i] = self.modalities[i][:, self.modalities[i].var['highly_variable']]
        else:
            print('Preprocessing method not supported.')
            return self
        self.preprocessed = True
        print('Preprocessing done.')
        return self

    def get_preprocessed_data(self):
        return self.preprocessed_data

    def normalize(self):
        assert self.preprocessed, 'Normalization must be conducted after preprocessing.'

        self.mean = np.concatenate([self.preprocessed_data['X_train'], self.preprocessed_data['X_test']], 0).mean()
        self.std = np.concatenate([self.preprocessed_data['X_train'], self.preprocessed_data['X_test']], 0).std()
        self.preprocessed_data['X_train'] = (self.preprocessed_data['X_train'] - self.mean) / self.std
        self.preprocessed_data['X_test'] = (self.preprocessed_data['X_test'] - self.mean) / self.std

        return self




import sys
import os.path as o
#use if running from dance/examples/spatial/cell_type_deconvo
root_path=o.abspath(o.join(o.dirname(sys.modules[__name__].__file__), "../../.."))
sys.path.append(root_path)


from dance.datasets.spatial import CellTypeDeconvoDataset
from dance.modules.spatial.cell_type_deconvo.spotlight import SPOTlight
import numpy as np
# get data 
dataset = CellTypeDeconvoDataset("toy2", data_dir="data/spatial")

X=np.array(dataset.data['ref_sc_count'])
Y=np.array(dataset.data['mix_count'])

#set cell-types 
cell_types = cell_types=np.array([np.repeat(i+1, X.shape[1]//3) for i in range(3)]).flatten()

#run deconvolution with SPOTlight (NMFReg + additional NNLS layer) - set rank for NMF dim red.
decon_model = SPOTlight(rank=10, bias=False)

#fit using reference scRNA X, with cell_type labels
decon_model.fit(X, Y, cell_types)
print(decon_model.P[:,:3])
#decon_model.P gives fitted cell-type proportions for Y 

#predict proportions at given spot 
prop_preds=decon_model.predict(Y[:,0])

#score model
print('Score:', decon_model.score(X, Y, cell_types))


import argparse
from pprint import pprint

from dance.datasets.singlemodality import CellTypeDataset
from dance.modules.single_modality.cell_type_annotation.svm import SVM

if __name__ == "__main__":
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--dense_dim", type=int, default=400, help="dim of PCA")
    parser.add_argument("--gpu", type=int, default=-1, help="GPU id, set to -1 for CPU")
    parser.add_argument("--map_path", default="map")
    parser.add_argument("--random_seed", type=int, default=10)
    parser.add_argument("--save_dir", default="result", help="Result directory")
    parser.add_argument("--species", default="mouse")
    parser.add_argument("--statistics_path", default="pretrained")
    parser.add_argument("--test_dataset", nargs="+", default=[299, 300], type=int, help="list of dataset id")
    parser.add_argument("--test_dir", default="test")
    parser.add_argument("--threshold", type=float, default=0, help="threshold to connect edges between cells and genes")
    parser.add_argument("--tissue", default="Testis")  # TODO: Add option for different tissue name for train/test
    parser.add_argument("--train_dataset", nargs="+", default=[2216], type=int, help="list of dataset id")
    parser.add_argument("--train_dir", default="train")

    params = parser.parse_args()
    pprint(vars(params))

    dataloader = CellTypeDataset(data_type="svm", random_seed=params.random_seed, train_dataset=params.train_dataset,
                                 test_dataset=params.test_dataset, species=params.species, tissue=params.tissue,
                                 train_dir=params.train_dir, test_dir=params.test_dir,
                                 statistics_path=params.statistics_path, map_path=params.map_path,
                                 threshold=params.threshold, gpu=params.gpu)
    dataloader = dataloader.load_data()

    model = SVM(params)
    model.fit(dataloader.svm_train_labels, dataloader.svm_train_cell_feat)
    prediction = model.predict(dataloader.svm_map_dict, dataloader.svm_id2label, dataloader.svm_test_label_dict,
                               dataloader.svm_test_feat_dict, dataloader.svm_test_cell_id_dict)
    accuracy = model.score(dataloader.svm_map_dict, dataloader.svm_id2label, dataloader.svm_test_label_dict,
                           dataloader.svm_test_feat_dict, dataloader.svm_test_cell_id_dict)

"""To reproduce SVM benchmarks, please refer to command lines belows:

Mouse Brain
$ python svm.py --species mouse --tissue Brain --train_dataset 753 3285 --test_dataset 2695

Mouse Spleen
$ python svm.py --species mouse --tissue Spleen --train_dataset 1970 --test_dataset 1759

Mouse Kidney
$ python svm.py --species mouse --tissue Kidney --train_dataset 4682 --test_dataset 203

"""



# Copyright 2022 DSE lab.  All rights reserved.

import time
import dgl
from dgl import nn as dglnn
from scipy.sparse import csc_matrix
import torch
import os
from collections import defaultdict
import pickle
from sklearn.decomposition import TruncatedSVD, PCA
import pandas as pd
import numpy as np
import numba
import scanpy as sc
from sklearn.preprocessing import normalize
import networkx as nx

import glob
import itertools
import random
from sklearn.neighbors import KDTree
import networkx as nx
from torch.nn import functional as F
from scipy import sparse as sp
from scipy.spatial import distance_matrix, minkowski_distance, distance
from sklearn.neighbors import kneighbors_graph
import dance.transforms.preprocess
import sklearn
from sklearn.metrics import pairwise_distances as pair


@numba.njit("f4(f4[:], f4[:])")
def euclid_dist(t1, t2):
    sum = 0
    for i in range(t1.shape[0]):
        sum += (t1[i] - t2[i]) ** 2
    return np.sqrt(sum)


@numba.njit("f4[:,:](f4[:,:])", parallel=True, nogil=True)
def pairwise_distance(X):
    n = X.shape[0]
    adj = np.empty((n, n), dtype=np.float32)
    for i in numba.prange(n):
        for j in numba.prange(n):
            adj[i][j] = euclid_dist(X[i], X[j])
    return adj


def csr_cosine_similarity(input_csr_matrix):
    similarity = input_csr_matrix * input_csr_matrix.T
    square_mag = similarity.diagonal()
    inv_square_mag = 1 / square_mag
    inv_square_mag[np.isinf(inv_square_mag)] = 0
    inv_mag = np.sqrt(inv_square_mag)
    res = similarity.multiply(inv_mag).T.multiply(inv_mag)
    return res.toarray()


def cosine_similarity_gene(input_matrix):
    from sklearn.metrics.pairwise import cosine_similarity
    res = cosine_similarity(input_matrix)
    res = np.abs(res)
    return res


def extract_color(x_pixel=None, y_pixel=None, image=None, beta=49):
    # beta to control the range of neighbourhood when calculate grey vale for one spot
    beta_half = round(beta / 2)
    g = []
    for i in range(len(x_pixel)):
        max_x = image.shape[0]
        max_y = image.shape[1]
        nbs = image[max(0, x_pixel[i] - beta_half):min(max_x, x_pixel[i] + beta_half + 1),
              max(0, y_pixel[i] - beta_half):min(max_y, y_pixel[i] + beta_half + 1)]
        g.append(np.mean(np.mean(nbs, axis=0), axis=0))
    c0, c1, c2 = [], [], []
    for i in g:
        c0.append(i[0])
        c1.append(i[1])
        c2.append(i[2])
    c0 = np.array(c0)
    c1 = np.array(c1)
    c2 = np.array(c2)
    c3 = (c0 * np.var(c0) + c1 * np.var(c1) + c2 * np.var(c2)) / (np.var(c0) + np.var(c1) + np.var(c2))
    return c3


def calculate_adj_matrix(x, y, x_pixel=None, y_pixel=None, image=None, beta=49, alpha=1, histology=True):
    # x,y,x_pixel, y_pixel are lists
    if histology:
        assert (x_pixel is not None) & (x_pixel is not None) & (image is not None)
        assert (len(x) == len(x_pixel)) & (len(y) == len(y_pixel))
        print("Calculateing adj matrix using histology image...")
        # beta to control the range of neighbourhood when calculate grey vale for one spot
        # alpha to control the color scale
        beta_half = round(beta / 2)
        g = []
        for i in range(len(x_pixel)):
            max_x = image.shape[0]
            max_y = image.shape[1]
            nbs = image[max(0, x_pixel[i] - beta_half):min(max_x, x_pixel[i] + beta_half + 1),
                  max(0, y_pixel[i] - beta_half):min(max_y, y_pixel[i] + beta_half + 1)]
            g.append(np.mean(np.mean(nbs, axis=0), axis=0))
        c0, c1, c2 = [], [], []
        for i in g:
            c0.append(i[0])
            c1.append(i[1])
            c2.append(i[2])
        c0 = np.array(c0)
        c1 = np.array(c1)
        c2 = np.array(c2)
        print("Var of c0,c1,c2 = ", np.var(c0), np.var(c1), np.var(c2))
        c3 = (c0 * np.var(c0) + c1 * np.var(c1) + c2 * np.var(c2)) / (np.var(c0) + np.var(c1) + np.var(c2))
        c4 = (c3 - np.mean(c3)) / np.std(c3)
        z_scale = np.max([np.std(x), np.std(y)]) * alpha
        z = c4 * z_scale
        z = z.tolist()
        print("Var of x,y,z = ", np.var(x), np.var(y), np.var(z))
        X = np.array([x, y, z]).T.astype(np.float32)
    else:
        print("Calculateing adj matrix using xy only...")
        X = np.array([x, y]).T.astype(np.float32)
    return pairwise_distance(X)


def construct_graph(x, y, x_pixel=None, y_pixel=None, image=None, beta=49, alpha=1, histology=True):
    return calculate_adj_matrix(x, y, x_pixel=x_pixel, y_pixel=y_pixel, image=image, beta=beta, alpha=alpha,
                                histology=histology)


def construct_enhanced_feature_graph(u, v, e, cell_node_features, enhance_graph=None, test=False, **kwargs):
    """Generate a feature-cell graph, enhanced with domain-knowledge (e.g. pathway).

    Parameters
    ----------
    u: torch.Tensor
        1-dimensional tensor. Cell node id of each cell-feature edge.
    v: torch.Tensor
        1-dimensional tensor. Feature node id of each cell-feature edge.
    e: torch.Tensor
        1-dimensional tensor. Weight of each cell-feature edge.
    cell_node_features: torch.Tensor
        1-dimensional or 2-dimensional tensor.  Node features for each cell node.
    enhance_graph: list[torch.Tensor]
        Node ids and edge weights of enhancement graph.

    Returns
    --------
    graph: DGLGraph
        The generated graph.
    """

    TRAIN_SIZE = kwargs['TRAIN_SIZE']
    FEATURE_SIZE = kwargs['FEATURE_SIZE']

    if enhance_graph is None:
        print('WARNING: Enhance graph disabled.')

    if kwargs['only_pathway'] and enhance_graph is not None:
        assert (kwargs['subtask'].find('rna') != -1)
        uu, vv, ee = enhance_graph

        graph_data = {
            ('feature', 'feature2cell', 'cell'): (v, u),
            ('feature', 'pathway', 'feature'): (uu, vv),
        }
        graph = dgl.heterograph(graph_data)

        if kwargs['inductive'] != 'trans':
            graph.nodes['cell'].data['id'] = cell_node_features[:TRAIN_SIZE] if not test else cell_node_features
        else:
            graph.nodes['cell'].data['id'] = cell_node_features

        graph.nodes['feature'].data['id'] = torch.arange(FEATURE_SIZE).long()
        graph.edges['feature2cell'].data['weight'] = e
        graph.edges['pathway'].data['weight'] = torch.tensor(ee).float()

    elif kwargs['no_pathway'] or kwargs['subtask'].find('rna') == -1 or enhance_graph is None:


        if kwargs['inductive'] == 'opt':
            print('Not supported.')
            # graph_data = {
            #     ('cell', 'cell2feature', 'feature'): (u, v) if not test else (
            #         u[:g.edges(etype='cell2feature')[0].shape[0]], v[:g.edges(etype='cell2feature')[0].shape[0]]),
            #     ('feature', 'feature2cell', 'cell'): (v, u),
            # }

        else:
            graph_data = {
                ('cell', 'cell2feature', 'feature'): (u, v),
                ('feature', 'feature2cell', 'cell'): (v, u),
            }

        graph = dgl.heterograph(graph_data)

        if kwargs['inductive'] != 'trans':
            graph.nodes['cell'].data['id'] = cell_node_features[:TRAIN_SIZE] if not test else cell_node_features
        else:
            graph.nodes['cell'].data['id'] = cell_node_features
        graph.nodes['feature'].data['id'] = torch.arange(FEATURE_SIZE).long()
        graph.edges['feature2cell'].data['weight'] = e
        graph.edges['cell2feature'].data['weight'] = e[:graph.edges(etype='cell2feature')[0].shape[0]]

    else:
        assert (kwargs['subtask'].find('rna') != -1)
        uu, vv, ee = enhance_graph

        if kwargs['inductive'] == 'opt':
            print("Not supported.")
            # graph_data = {
            #     ('cell', 'cell2feature', 'feature'): (u, v) if not test else (
            #         u[:g.edges(etype='cell2feature')[0].shape[0]], v[:g.edges(etype='cell2feature')[0].shape[0]]),
            #     ('feature', 'feature2cell', 'cell'): (v, u),
            #     ('feature', 'pathway', 'feature'): (uu, vv),
            # }
        else:
            graph_data = {
                ('cell', 'cell2feature', 'feature'): (u, v),
                ('feature', 'feature2cell', 'cell'): (v, u),
                ('feature', 'pathway', 'feature'): (uu, vv),
            }
        graph = dgl.heterograph(graph_data)

        if kwargs['inductive'] != 'trans':
            graph.nodes['cell'].data['id'] = cell_node_features[:TRAIN_SIZE] if not test else cell_node_features
        else:
            graph.nodes['cell'].data['id'] = cell_node_features
        graph.nodes['feature'].data['id'] = torch.arange(FEATURE_SIZE).long()
        graph.edges['feature2cell'].data['weight'] = e
        graph.edges['cell2feature'].data['weight'] = e[:graph.edges(etype='cell2feature')[0].shape[0]]
        graph.edges['pathway'].data['weight'] = torch.tensor(ee).float()

    return graph


# TODO: haven't explained extra kwargs
def construct_pathway_graph(gex_data, **kwargs):
    """Generate nodes, edges and edge weights for pathway dataset.

    Parameters
    ----------
    gex_data: anndata.AnnData
        Gene data, contains feature matrix (.X) and feature names (.var['feature_types']).

    Returns
    --------
    uu: list[int]
        Predecessor node id of each edge.
    vv: list[int]
        Successor node id of each edge.
    ee: list[float]
        Edge weight of each edge.
    """

    pww = kwargs['pathway_weight']
    npw = kwargs['no_pathway']
    subtask = kwargs['subtask']
    pw_path = kwargs['pathway_path']
    uu = []
    vv = []
    ee = []

    assert (not npw)

    pk_path = f'pw_{subtask}_{pww}.pkl'
    #     pk_path = f'pw_{subtask}_{pww}.pkl'
    if os.path.exists(pk_path):
        print(
            'WARNING: Pathway file exist. Load pickle file by default. Auguments "--pathway_weight" and "--pathway_path" will not take effect.')
        uu, vv, ee = pickle.load(open(pk_path, 'rb'))
    else:
        # Load Original Pathway File
        with open(pw_path + '.entrez.gmt') as gmt:
            gene_list = gmt.read().split()

        gene_sets_entrez = defaultdict(list)
        indicator = 0
        for ele in gene_list:
            if not ele.isnumeric() and indicator == 1:
                indicator = 0
                continue
            if not ele.isnumeric() and indicator == 0:
                indicator = 1
                gene_set_name = ele
            else:
                gene_sets_entrez[gene_set_name].append(ele)

        with open(pw_path + '.symbols.gmt') as gmt:
            gene_list = gmt.read().split()

        gene_sets_symbols = defaultdict(list)

        for ele in gene_list:
            if ele in gene_sets_entrez:
                gene_set_name = ele
            elif not ele.startswith('http://'):
                gene_sets_symbols[gene_set_name].append(ele)

        pw = [i[1] for i in gene_sets_symbols.items()]

        # Generate New Pathway Data
        counter = 0
        total = 0
        feature_index = gex_data.var['feature_types'].index.tolist()
        gex_features = gex_data.X
        new_pw = []
        for i in pw:
            new_pw.append([])
            for j in i:
                if j in feature_index:
                    new_pw[-1].append(feature_index.index(j))

        if pww == 'cos':
            for i in new_pw:
                for j in i:
                    for k in i:
                        if j != k:
                            uu.append(j)
                            vv.append(k)
                            sj = np.sqrt(
                                np.dot(gex_features[:, j].toarray().T, gex_features[:, j].toarray()).item())
                            sk = np.sqrt(
                                np.dot(gex_features[:, k].toarray().T, gex_features[:, k].toarray()).item())
                            jk = np.dot(gex_features[:, j].toarray().T, gex_features[:, k].toarray())
                            cossim = jk / sj / sk
                            ee.append(cossim.item())
        elif pww == 'one':
            for i in new_pw:
                for j in i:
                    for k in i:
                        if j != k:
                            uu.append(j)
                            vv.append(k)
                            ee.append(1.)
        elif pww == 'pearson':
            corr = np.corrcoef(gex_features.toarray().T)
            for i in new_pw:
                for j in i:
                    for k in i:
                        if j != k:
                            uu.append(j)
                            vv.append(k)
                            ee.append(corr[j][k])

        pickle.dump([uu, vv, ee], open(pk_path, 'wb'))

    # Apply Threshold
    pwth = kwargs['pathway_threshold']
    nu = []
    nv = []
    ne = []

    for i in range(len(uu)):
        if abs(ee[i]) > pwth:
            ne.append(ee[i])
            nu.append(uu[i])
            nv.append(vv[i])
    uu, vv, ee = nu, nv, ne

    return uu, vv, ee


def construct_basic_feature_graph(feature_mod1, feature_mod1_test=None, bf_input=None, device='cuda'):
    input_train_mod1 = csc_matrix(feature_mod1)

    """Generate a feature-cell graph, enhanced with domain-knowledge (e.g. pathway).

    Parameters
    ----------
    feature_mod1 : torch.Tensor
        Features of input modality.
    feature_mod1_test : torch.Tensor optional
        Features of input modality of testing samples.
    bf_input : torch.Tensor optional
        Batch features, by default to be None. If undefined, cell nodes would not have initialization.
    device : str optional
        The device where the graph would be generated, by default to be 'cuda'.

    Returns
    --------
    g: DGLGraph
        The generated graph.
    """

    if feature_mod1_test is not None:
        input_test_mod1 = csc_matrix(feature_mod1_test)
        assert (input_test_mod1.shape[1] == input_train_mod1.shape[1])

        u = torch.from_numpy(np.concatenate([np.array(t.nonzero()[0] + i) for i, t in enumerate(input_train_mod1)] + \
                                            [np.array(t.nonzero()[0] + i + input_train_mod1.shape[0]) for i, t in
                                             enumerate(input_test_mod1)], axis=0))
        v = torch.from_numpy(np.concatenate([np.array(t.nonzero()[1]) for t in input_train_mod1] + \
                                            [np.array(t.nonzero()[1]) for t in input_test_mod1], axis=0))
        sample_size = input_train_mod1.shape[0] + input_test_mod1.shape[0]
        weights = torch.from_numpy(
            np.concatenate([input_train_mod1.tocsr().data, input_test_mod1.tocsr().data], axis=0)).float()

    else:
        u = torch.from_numpy(
            np.concatenate([np.array(t.nonzero()[0] + i) for i, t in enumerate(input_train_mod1)], axis=0))
        v = torch.from_numpy(np.concatenate([np.array(t.nonzero()[1]) for t in input_train_mod1], axis=0))
        sample_size = input_train_mod1.shape[0]
        weights = torch.from_numpy(np.concatenate([input_train_mod1.tocsr().data], axis=0)).float()

    graph_data = {
        ('cell', 'cell2feature', 'feature'): (u, v),
        ('feature', 'feature2cell', 'cell'): (v, u),
    }
    g = dgl.heterograph(graph_data)

    if bf_input:
        g.nodes['cell'].data['bf'] = gen_batch_features(bf_input)
    else:
        g.nodes['cell'].data['bf'] = torch.zeros(sample_size).float()

    g.nodes['cell'].data['id'] = torch.zeros(sample_size).long() #torch.arange(sample_size).long()
    #     g.nodes['cell'].data['source'] =
    g.nodes['feature'].data['id'] = torch.arange(input_train_mod1.shape[1]).long()
    g.edges['cell2feature'].data['weight'] = g.edges['feature2cell'].data['weight'] = weights

    g = g.to(device)
    return g


def gen_batch_features(ad_inputs):
    """Generate statistical features for each batch in the input data, and assign batch features to each cell.
    This function returns batch features for each cell in all the input sub-datasets.

    Parameters
    ----------
    ad_inputs: list[anndata.AnnData]
        A list of AnnData object, each contains a sub-dataset.

    Returns
    --------
    batch_features: torch.Tensor
        A batch_features matrix, each row refers to one cell from the datasets. The matrix can be directly used as the
        node features of cell nodes.
    """

    cells = []
    columns = ['cell_mean', 'cell_std', 'nonzero_25%', 'nonzero_50%', 'nonzero_75%', 'nonzero_max', 'nonzero_count',
               'nonzero_mean', 'nonzero_std', 'batch']

    assert len(ad_inputs) < 10, "WARNING: Input of gen_bf_features should be a list of AnnData objects."

    for ad_input in ad_inputs:
        bcl = list(ad_input.obs['batch'])
        print(set(bcl))
        for i, cell in enumerate(ad_input.X):
            cell = cell.toarray()
            nz = cell[np.nonzero(cell)]
            if len(nz)==0:
                print('Error: one cell contains all zero features.')
                exit()
            cells.append([cell.mean(), cell.std(), np.percentile(nz, 25), np.percentile(nz, 50), np.percentile(nz, 75),
                          cell.max(), len(nz) / 1000, nz.mean(), nz.std(), bcl[i]])

    cell_features = pd.DataFrame(cells, columns=columns)
    batch_source = cell_features.groupby('batch').mean().reset_index()
    batch_list = batch_source.batch.tolist()
    batch_source = batch_source.drop('batch', axis=1).to_numpy().tolist()
    b2i = dict(zip(batch_list, range(len(batch_list))))
    batch_features = []

    for ad_input in ad_inputs:
        for b in ad_input.obs['batch']:
            batch_features.append(batch_source[b2i[b]])

    batch_features = torch.tensor(batch_features).float()

    return batch_features


def construct_modality_prediction_graph(dataset, **kwargs):
    """Construct the cell-feature graph object for modality prediction task, based on the input dataset.

    Parameters
    ----------
    dataset: datasets.multimodality.ModalityPredictionDataset
        The input dataset, typically includes four input AnnData sub-datasets, which are train_mod1, train_mod2,
        test_mod1 and test_mod2 respectively.

    Returns
    --------
    g: DGLGraph
        The generated graph.
    """

    train_mod1 = dataset.modalities[0]
    input_train_mod1 = dataset.sparse_features()[0]
    if kwargs['inductive'] == 'trans':
        input_test_mod1 = dataset.sparse_features()[2]

    CELL_SIZE = kwargs['CELL_SIZE']
    TRAIN_SIZE = kwargs['TRAIN_SIZE']

    if kwargs['cell_init'] == 'none':
        cell_node_features = torch.ones(CELL_SIZE).long()
    elif kwargs['cell_init'] == 'pca':
        embedder_mod1 = TruncatedSVD(n_components=100)
        X_train_np = embedder_mod1.fit_transform(input_train_mod1.toarray())
        X_test_np = embedder_mod1.transform(input_test_mod1.toarray())
        cell_node_features = torch.cat([torch.from_numpy(X_train_np), torch.from_numpy(X_test_np)], 0).float()
    if (not kwargs['no_pathway']) and (kwargs['subtask'].find('rna') != -1):
        enhance_graph = construct_pathway_graph(train_mod1, **kwargs)
    else:
        enhance_graph = None

    if kwargs['inductive'] != 'trans':
        u = torch.from_numpy(
            np.concatenate([np.array(t.nonzero()[0] + i) for i, t in enumerate(input_train_mod1)], axis=0))
        v = torch.from_numpy(np.concatenate([np.array(t.nonzero()[1]) for t in input_train_mod1], axis=0))
        e = torch.from_numpy(input_train_mod1.tocsr().data).float()
        g = construct_enhanced_feature_graph(u, v, e, cell_node_features, enhance_graph, **kwargs)

        u = torch.from_numpy(np.concatenate([np.array(t.nonzero()[0] + i) for i, t in enumerate(input_train_mod1)] + \
                                            [np.array(t.nonzero()[0] + i + TRAIN_SIZE) for i, t in
                                             enumerate(input_test_mod1)], axis=0))
        v = torch.from_numpy(np.concatenate([np.array(t.nonzero()[1]) for t in input_train_mod1] + \
                                            [np.array(t.nonzero()[1]) for t in input_test_mod1], axis=0))
        e = torch.from_numpy(
            np.concatenate([input_train_mod1.tocsr().data, input_test_mod1.tocsr().data], axis=0)).float()
        gtest = construct_enhanced_feature_graph(u, v, e, cell_node_features, enhance_graph, test=True, **kwargs)
        return g, gtest

    else:
        u = torch.from_numpy(np.concatenate([np.array(t.nonzero()[0] + i) for i, t in enumerate(input_train_mod1)] + \
                                            [np.array(t.nonzero()[0] + i + TRAIN_SIZE) for i, t in
                                             enumerate(input_test_mod1)], axis=0))
        v = torch.from_numpy(np.concatenate([np.array(t.nonzero()[1]) for t in input_train_mod1] + \
                                            [np.array(t.nonzero()[1]) for t in input_test_mod1], axis=0))
        e = torch.from_numpy(
            np.concatenate([input_train_mod1.tocsr().data, input_test_mod1.tocsr().data], axis=0)).float()
        g = construct_enhanced_feature_graph(u, v, e, cell_node_features, enhance_graph, **kwargs)

        return g


def make_graph(X, Y=None, threshold=0, dense_dim=100, gene_data={},
               normalize_weights="log_per_cell", nb_edges=1,
               node_features="scale", same_edge_values=False,
               edge_norm=True):
    """Create DGL graph for graph-sc

    Parameters
    ----------
    X :
        input cell-gene features.
    Y : list optional
        true labels.
    threshold : int optional
        minimum value of selected feature.
    dense_dim : int optional
        dense dimension for PCA.
    gene_data : dict optional
        external gene data.
    normalize_weights : str optional
        weights normalization method.
    nb_edges : float, optional
        proportion of edges selected.
    node_features : str optional
        type of node features.
    same_edge_values : bool optional
        set identical edge value or not.
    edge_norm : bool optional
        perform edge normalization or not.

    Returns
    -------
    graph :
        constructed dgl graph.

    """
    num_genes = X.shape[1]

    graph = dgl.DGLGraph()
    gene_ids = torch.arange(X.shape[1], dtype=torch.int32).unsqueeze(-1)
    graph.add_nodes(num_genes, {'id': gene_ids})

    row_idx, gene_idx = np.nonzero(X > threshold)  # intra-dataset index

    if normalize_weights == "none":
        X1 = X
    if normalize_weights == "log_per_cell":
        X1 = np.log1p(X)
        X1 = X1 / (np.sum(X1, axis=1, keepdims=True) + 1e-6)

    if normalize_weights == "per_cell":
        X1 = X / (np.sum(X, axis=1, keepdims=True) + 1e-6)

    non_zeros = X1[(row_idx, gene_idx)]  # non-zero values

    cell_idx = row_idx + graph.number_of_nodes()  # cell_index
    cell_nodes = torch.tensor([-1] * len(X), dtype=torch.int32).unsqueeze(-1)

    graph.add_nodes(len(cell_nodes), {'id': cell_nodes})
    if nb_edges > 0:
        edge_ids = np.argsort(non_zeros)[::-1]
    else:
        edge_ids = np.argsort(non_zeros)
        nb_edges = abs(nb_edges)
        print(f"selecting weakest edges {int(len(edge_ids) * nb_edges)}")
    edge_ids = edge_ids[:int(len(edge_ids) * nb_edges)]
    cell_idx = cell_idx[edge_ids]
    gene_idx = gene_idx[edge_ids]
    non_zeros = non_zeros[edge_ids]

    if same_edge_values:
        graph.add_edges(
            gene_idx, cell_idx, {
                'weight':
                    torch.tensor(np.ones_like(non_zeros),
                                 dtype=torch.float32).unsqueeze(1)
            })
    else:
        graph.add_edges(
            gene_idx, cell_idx, {
                'weight':
                    torch.tensor(non_zeros, dtype=torch.float32).unsqueeze(1)
            })

    if node_features == "scale":
        nX = ((X1 - np.mean(X1, axis=0)) / np.std(X1, axis=0))
        gene_feat = PCA(dense_dim, random_state=1).fit_transform(
            nX.T).astype(float)
        cell_feat = X1.dot(gene_feat).astype(float)
    if node_features == "scale_by_cell":
        nX = ((X1 - np.mean(X1, axis=0)) / np.std(X1, axis=0))
        cell_feat = PCA(dense_dim, random_state=1).fit_transform(
            nX).astype(float)
        gene_feat = X1.T.dot(cell_feat).astype(float)
    if node_features == "none":
        gene_feat = PCA(dense_dim, random_state=1).fit_transform(
            X1.T).astype(float)
        cell_feat = X1.dot(gene_feat).astype(float)

    graph.ndata['features'] = torch.cat([torch.from_numpy(gene_feat),
                                         torch.from_numpy(cell_feat)],
                                        dim=0).type(torch.float)

    graph.ndata['order'] = torch.tensor([-1] * num_genes + list(np.arange(len(X))),
                                        dtype=torch.long)  # [gene_num+train_num]
    if Y is not None:
        graph.ndata['label'] = torch.tensor([-1] * num_genes + list(np.array(Y).astype(int)),
                                            dtype=torch.long)  # [gene_num+train_num]
    else:
        graph.ndata['label'] = torch.tensor(
            [-1] * num_genes + [np.nan] * len(X))
    nb_edges = graph.num_edges()

    if len(gene_data) != 0 and len(gene_data['gene1']) > 0:
        graph = external_data_connections(
            graph, gene_data, X, gene_idx, cell_idx)
    in_degrees = graph.in_degrees()
    # Edge normalization
    if edge_norm:
        for i in range(graph.number_of_nodes()):
            src, dst, in_edge_id = graph.in_edges(i, form='all')
            if src.shape[0] == 0:
                continue
            edge_w = graph.edata['weight'][in_edge_id]
            graph.edata['weight'][in_edge_id] = in_degrees[i] * edge_w / torch.sum(
                edge_w)

    graph.add_edges(
        graph.nodes(), graph.nodes(), {
            'weight':
                torch.ones(graph.number_of_nodes(),
                           dtype=torch.float).unsqueeze(1)
        })
    return graph

def external_data_connections(graph, gene_data, X, gene_idx, cell_idx):
    """Add external data to graph

    Parameters
    ----------
    graph :
        dgl graph.
    gene_data : dict optional
        external gene data.
    X :
        input cell-gene features.
    gene_idx : list
        index of gene.
    cell_idx : list
        index of cell.

    Returns
    -------
    graph :
        constructed dgl graph.

    """
    num_genes = X.shape[1]
    initial_nb_edges = graph.num_edges()
    if gene_data.get("single_layer", False) == True:
        sel_cell_idx = np.argsort((X > 0).sum(axis=1))[
            :int(len(X) * gene_data["select_cells"])]
        sel_cell_idx += X.shape[1]

        normalized_w_values = graph.edata["weight"].numpy().reshape(-1)
        exclude_high_genes = []

        for cell_id in tqdm(sel_cell_idx):
            all_existing_genes = gene_idx[np.where(cell_idx == cell_id)[0]]
            existing_genes_w = normalized_w_values[np.where(cell_idx == cell_id)[
                0]]
            keep_idx = np.where(
                ~np.isin(all_existing_genes, exclude_high_genes))[0]
            existing_genes = all_existing_genes[keep_idx]
            existing_genes_w = existing_genes_w[keep_idx]
            # select random genes
            strond_id = np.random.choice(np.arange(len(existing_genes)),
                                         gene_data["select_genes_threshold"], replace=False)

            existing_genes = existing_genes[strond_id]
            existing_genes_w = existing_genes_w[strond_id]

            for i, g in enumerate(existing_genes):
                correlated_ids = np.where(gene_data['gene2'] == g)[0]
                correlated_genes = gene_data['gene1'][correlated_ids]
                ii = np.where(
                    ~np.isin(correlated_genes, all_existing_genes))[0]
                correlated_ids = correlated_ids[ii]
                correlated_genes = correlated_genes[ii]
                correlated_weights = gene_data['gene_weights'][correlated_ids]
                if len(correlated_genes) > 0:
                    best_id = np.argsort(correlated_weights)[
                        ::-1][:gene_data["nb_correlated_genes"]]
                    graph.add_edges(
                        correlated_genes[best_id], [cell_id] * len(best_id), {
                            'weight':
                            torch.tensor(existing_genes_w[i]*np.ones_like(best_id),
                                         dtype=torch.float32).unsqueeze(1)
                        })
    else:

        weights = torch.from_numpy(gene_data['gene_weights'].astype(
            np.float32) * gene_data['weight']).unsqueeze(1)
        graph.add_edges(gene_data['gene1'], gene_data['gene2'],
                        {'weight': weights})
        print("Adding gene to gene relations", gene_data['gene_weights'].shape, gene_data['weight'],
              weights.max(), weights.min())
    gene_data["extra_edges"] = (
        graph.num_edges() - initial_nb_edges)/graph.num_edges()
    return graph

def get_adj(count, k=15, pca_dim=50, mode="connectivity"):
    '''Conctruct adjacency matrix for scTAG

    Parameters
    ----------
    count :
        input cell-gene features.
    k : int optional
        number of neighbors for each sample in k-neighbors graph.
    pca_dim : int optional
        number of components in PCA.
    mode : str optional
        type of returned adjacency matrix.

    Returns
    -------
    pred : list
        prediction of leiden.
    '''
    if pca_dim:
        countp = PCA(n_components=pca_dim).fit_transform(count)
    else:
        countp = count
    A = kneighbors_graph(countp, k, mode=mode, metric="euclidean", include_self=True)
    adj = A.toarray()
    normalized_D = degree_power(adj, -0.5)
    adj_n = normalized_D.dot(adj).dot(normalized_D)

    return adj, adj_n


def degree_power(A, k):
    degrees = np.power(np.array(A.sum(1)), k).flatten()
    degrees[np.isinf(degrees)] = 0.
    if sp.issparse(A):
        D = sp.diags(degrees)
    else:
        D = np.diag(degrees)
    return D


############################
# For cell type deconvolution
############################

def kNN(data, k, query=None):
    tree = KDTree(data)
    if query is None:
        query = data
    dist, ind = tree.query(query, k)
    return dist, ind


# ' @param cell_embedding : pandas data frame
def KNN(cell_embedding, spots1, spots2, k):
    embedding_spots1 = cell_embedding.loc[spots1,]
    embedding_spots2 = cell_embedding.loc[spots2,]
    nnaa = kNN(embedding_spots1, k=k + 1)
    nnbb = kNN(embedding_spots2, k=k + 1)
    nnab = kNN(data=embedding_spots2, k=k, query=embedding_spots1)
    nnba = kNN(data=embedding_spots1, k=k, query=embedding_spots2)
    return nnaa, nnab, nnba, nnbb, spots1, spots2


def MNN(neighbors, colnames, num):
    max_nn = np.array([neighbors[1][1].shape[1], neighbors[2][1].shape[1]])
    if ((num > max_nn).any()):
        num = np.min(max_nn)
        # convert cell name to neighbor index
    spots1 = colnames
    spots2 = colnames
    nn_spots1 = neighbors[4]
    nn_spots2 = neighbors[5]
    cell1_index = [
        list(nn_spots1).index(i) for i in spots1 if (nn_spots1 == i).any()
    ]
    cell2_index = [
        list(nn_spots2).index(i) for i in spots2 if (nn_spots2 == i).any()
    ]
    ncell = range(neighbors[1][1].shape[0])
    ncell = np.array(ncell)[np.in1d(ncell, cell1_index)]
    # initialize a list
    mnn_cell1 = [None] * (len(ncell) * 5)
    mnn_cell2 = [None] * (len(ncell) * 5)
    idx = -1
    for cell in ncell:
        neighbors_ab = neighbors[1][1][cell, 0:5]
        mutual_neighbors = np.where(
            neighbors[2][1][neighbors_ab, 0:5] == cell)[0]
        for i in neighbors_ab[mutual_neighbors]:
            idx = idx + 1
            mnn_cell1[idx] = cell
            mnn_cell2[idx] = i
    mnn_cell1 = mnn_cell1[0:(idx + 1)]
    mnn_cell2 = mnn_cell2[0:(idx + 1)]
    import pandas as pd
    mnns = pd.DataFrame(np.column_stack((mnn_cell1, mnn_cell2)))
    mnns.columns = ['spot1', 'spot2']
    return mnns


def filterEdge(edges, neighbors, mats, features, k_filter):
    nn_spots1 = neighbors[4]
    nn_spots2 = neighbors[5]
    mat1 = mats.loc[features, nn_spots1].transpose()
    mat2 = mats.loc[features, nn_spots2].transpose()
    cn_data1 = dance.transforms.preprocess.l2norm(mat1)
    cn_data2 = dance.transforms.preprocess.l2norm(mat2)
    nn = kNN(data=cn_data2.loc[nn_spots2,],
             query=cn_data1.loc[nn_spots1,],
             k=k_filter)
    position = [
        np.where(
            edges.loc[:, "spot2"][x] == nn[1][edges.loc[:, 'spot1'][x],])[0]
        for x in range(edges.shape[0])
    ]
    nps = np.concatenate(position, axis=0)
    fedge = edges.iloc[nps,]
    return (fedge)


def stLinkGraph(  # count_list,
        # norm_list,
        scale_list,
        # features,
        combine,
        k_filter=200, num_cc=30):
    all_edges = []
    for row in combine:
        i = row[0]
        j = row[1]
        # counts1 = count_list[i]
        # counts2 = count_list[j]
        # norm_data1 = norm_list[i]
        # norm_data2 = norm_list[j]
        scale_data1 = scale_list[i]
        scale_data2 = scale_list[j]

        cell_embedding, loading = dance.transforms.preprocess.ccaEmbed(scale_data1,
                                                                       scale_data2,
                                                                       num_cc=num_cc)
        norm_embedding = dance.transforms.preprocess.l2norm(mat=cell_embedding[0])
        spots1 = scale_data1.columns
        spots2 = scale_data2.columns
        neighbor = KNN(cell_embedding=norm_embedding,
                       spots1=spots1,
                       spots2=spots2,
                       k=30)
        mnn_edges = MNN(neighbors=neighbor,
                        colnames=cell_embedding[0].index,
                        num=5)
        select_genes = dance.transforms.preprocess.selectTopGenes(Loadings=loading,
                                                                  dims=range(num_cc),
                                                                  DimGenes=100,
                                                                  maxGenes=200)
        Mat = pd.concat([scale_data1, scale_data2], axis=1)
        final_edges = filterEdge(edges=mnn_edges,
                                 neighbors=neighbor,
                                 mats=Mat,
                                 features=select_genes,
                                 k_filter=k_filter)
        final_edges['Dataset1'] = [i + 1] * final_edges.shape[0]
        final_edges['Dataset2'] = [j + 1] * final_edges.shape[0]
        all_edges.append(final_edges)
    return all_edges



def stLinkGraphConstruct(st_scale, st_label, k_filter):
    N = len(st_scale)
    if (N == 1):
        combine = pd.Series([(0, 0)])
    else:
        combin = list(itertools.product(list(range(N)), list(range(N))))
        index = [i for i, x in enumerate([i[0] < i[1] for i in combin]) if x]
        combine = pd.Series(combin)[index]

    link1 = stLinkGraph(  # count_list=count_list,
        # norm_list=norm_list,
        scale_list=st_scale,
        # features=features,
        combine=combine, 
        k_filter=k_filter)

    # ' ---- input data for link grpah 2 -----
    # files1 = glob.glob(path0 + "/ST_count/*.csv")
    # files1.sort()
    # tem_count = pd.read_csv(files1[1], index_col=0)
    tem_scale = st_scale[1]
    #tem_scale.columns = tem_scale.columns.str.replace("mixt_", "rept_")

    tem_label = st_label[1]
    #tem_label.columns = tem_label.columns.str.replace("mixt_", "rept_")

    # count_list2 = [count_list[1],tem_count]
    # norm_list2 = [norm_list[1], tem_norm]
    scale_list2 = [st_scale[1], tem_scale]

    link2 = stLinkGraph(  # count_list=count_list2,
        # norm_list=norm_list2,
        scale_list=scale_list2,
        # features=features,
        combine=combine,
        k_filter=k_filter)

    graph1 = link1[0].iloc[:, 0:2].reset_index()
    graph1 = graph1.iloc[:, 1:3]
    # graph1.to_csv('./Datadir/Linked_graph1.csv')

    graph2 = link2[0].iloc[:, 0:2].reset_index()
    graph2 = graph2.iloc[:, 1:3]
    # graph2.to_csv('./Datadir/Linked_graph2.csv')

    # label1 = label_list[0]
    # label1.to_csv('./Datadir/Pseudo_Label1.csv', index=False)

    # label2 = label_list[1]
    # label2.to_csv('./Datadir/Real_Label2.csv', index=False)
    return ([graph1, graph2])



def stAdjConstruct(st_scale, st_label, adj_data, k_filter=1):
    
    st1 = pd.DataFrame(st_scale[0].X,index=st_scale[0].obs.index,
                columns=st_scale[0].var.index)
    st2 = pd.DataFrame(st_scale[1].X,index=st_scale[1].obs.index,
                    columns=st_scale[1].var.index)
    
    st_scale = [st1.transpose(), st2.transpose()]
    # ' construct adjacent matrix
    graphs = stLinkGraphConstruct(st_scale, st_label, k_filter)
    id_graph1 = graphs[0].copy()
    
    data_train1, data_val1, data_test1,labels, lab_data2 = adj_data[0], adj_data[1], adj_data[2], adj_data[3], adj_data[4]
    
    # ' map index
    fake1 = np.array([-1] * len(lab_data2.index))
    index1 = np.concatenate((data_train1.index, fake1, data_val1.index,
                             data_test1.index)).flatten()
    # ' (feature_data.index==index1).all()
    fake2 = np.array([-1] * len(data_train1))
    fake3 = np.array([-1] * (len(data_val1) + len(data_test1)))
    find1 = np.concatenate((fake2, np.array(lab_data2.index), fake3)).flatten()

    row1 = [np.where(find1 == id_graph1.iloc[i, 1])[0][0]
            for i in range(len(id_graph1))
            ]
    col1 = [np.where(index1 == id_graph1.iloc[i, 0])[0][0]
            for i in range(len(id_graph1))
            ]
    adj = defaultdict(
        list)  # default value of int is 0
    for i in range(len(labels)):
        adj[i].append(i)
    for i in range(len(row1)):
        adj[row1[i]].append(col1[i])
        adj[col1[i]].append(row1[i])

    adj = nx.adjacency_matrix(nx.from_dict_of_lists(adj))
    return adj


############################
#          scDSC           #
############################
def construct_graph_sc(fname, features, label, method, topk):
    """Graph construction function for scDSC

    Parameters
    ----------
    fname : str
        file name to save graph.
    features :
        input cell-gene features.
    label : list
        label of cells.
    method : str
        method to construct graph adjacency matrix.
    topk : int
        number of highly variable genes.

    Returns
    -------
    None.

    """
    num = len(label)
    if topk == None:
        topk = 0
    dist = None
    # Several methods of calculating the similarity relationship between samples i and j (similarity matrix Sij)
    if method == 'heat':
        dist = -0.5 * pair(features, metric='manhattan') ** 2
        dist = np.exp(dist)

    elif method == 'cos':
        features[features > 0] = 1
        dist = np.dot(features, features.T)

    elif method == 'ncos':
        features[features > 0] = 1
        features = normalize(features, axis=1, norm='l1')
        dist = np.dot(features, features.T)

    elif method == 'p':
        y = features.T - np.mean(features.T)
        features = features - np.mean(features)
        dist = np.dot(features, features.T) / (np.linalg.norm(features) * np.linalg.norm(y))

    if topk:
        inds = []
        for i in range(dist.shape[0]):
            ind = np.argpartition(dist[i, :], -(topk + 1))[-(topk + 1):]
            inds.append(ind)
    else:
        inds = dist

    f = open(fname, 'w')
    counter = 0
    A = np.zeros_like(dist)
    for i, v in enumerate(inds):
        mutual_knn = False
        for vv in v:
            if vv == i:
                pass
            else:
                if label[vv] != label[i]:
                    counter += 1
                f.write('{} {}\n'.format(i, vv))
    f.close()


def basic_feature_propagation(dataset, layers, transformed = True, device='cuda'):
    if transformed:
        g1 = construct_basic_feature_graph(dataset.preprocessed_features['mod1_train'], dataset.preprocessed_features['mod1_test'], device=device)
        g2 = construct_basic_feature_graph(dataset.preprocessed_features['mod2_train'], dataset.preprocessed_features['mod2_test'], device=device)
    else:
        g1 = construct_basic_feature_graph(dataset.sparse_features()[0], dataset.sparse_features()[2], device=device)
        g2 = construct_basic_feature_graph(dataset.sparse_features()[1], dataset.sparse_features()[3], device=device)

    hcell_mod1 = basic_feature_graph_propagation(g1, layers)
    hcell_mod2 = basic_feature_graph_propagation(g2, layers)

    return hcell_mod1, hcell_mod2


def basic_feature_graph_propagation(g, layers=3, alpha=0.5, beta=0.5, cell_init=None, feature_init='id', device='cuda', verbose=True):

    assert layers>2, 'Less than two feature graph propagation layers is equivalent to original features.'

    gconv = dglnn.HeteroGraphConv({
                    'cell2feature' : dglnn.GraphConv(in_feats=0, out_feats=0, norm='none', weight=False, bias=False),
                    'feature2cell' : dglnn.GraphConv(in_feats=0, out_feats=0, norm='none', weight=False, bias=False),
                },
            aggregate='sum')

    if feature_init is None:
        feature_X = torch.zeros((g.nodes('feature').shape[0], g.srcdata[cell_init]['cell'].shape[1])).to(device)
    elif feature_init == 'id':
        feature_X = F.one_hot(g.srcdata['id']['feature']).float().to(device)


    if cell_init is None:
        cell_X = torch.zeros(g.nodes('cell').shape[0], feature_X.shape[1]).to(device)
    else:
        cell_X = g.srcdata[cell_init]['cell']

    h = {'feature': feature_X,
         'cell': cell_X}
    hcell = []
    for i in range(layers):
        h1 = gconv(g, h, mod_kwargs={'cell2feature':{'edge_weight': g.edges['cell2feature'].data['weight']},
                                     'feature2cell':{'edge_weight': g.edges['feature2cell'].data['weight']}})
        if verbose: print(i, 'cell', h['cell'].abs().mean(), h1['cell'].abs().mean())
        if verbose: print(i, 'feature', h['feature'].abs().mean(), h1['feature'].abs().mean())

        h1['feature'] = (h1['feature'] - h1['feature'].mean())/(h1['feature'].std() if h1['feature'].mean()!=0 else 1)
        h1['cell'] = (h1['cell'] - h1['cell'].mean())/(h1['cell'].std() if h1['cell'].mean()!=0 else 1)

        h = {'feature': h['feature']*alpha + h1['feature']*(1-alpha),
             'cell': h['cell']*beta + h1['cell']*(1-beta)}

        h['feature'] = (h['feature'] - h['feature'].mean())/h['feature'].std()
        h['cell'] = (h['cell'] - h['cell'].mean())/h['cell'].std()

        hcell.append(h['cell'])

    if verbose: print(hcell[-1].abs().mean())

    return hcell[1:]

##############################
# neighbor_graph for stlearn #
############################## 

def neighbors_get_adj(adata, 
            n_neighbors=10, 
            n_pcs=10, 
            n_jobs=1, 
            use_rep=None,
            knn=True,
            random_state=None,
            method='umap',
            metric='euclidean',
            metric_kwargs={},
            copy=False,
            obsp =None, 
            neighbors_key=None ):

    sc.pp.neighbors(
        adata,
        n_neighbors=n_neighbors,
        n_pcs=n_pcs,
        use_rep=use_rep,
        knn=knn,
        random_state=random_state,
        method=method,
        metric=metric,
        metric_kwds=metric_kwargs,
        copy=copy,
    )

    choose_graph = getattr(sc._utils,"_choose_graph",None)
    adjacency = choose_graph(adata, obsp, neighbors_key)

    print("Created k-Nearest-Neighbor graph in adata.uns['neighbors'] ")
    return adjacency


##### scGNN create adjacency, likely much overlap with above functions, nested function defs to avoid possible namespace conflicts

def scGNNgenerateAdj(featureMatrix, graphType='KNNgraph', para = None, parallelLimit = 0, adjTag = True ):
    """
    Generating edgeList
    """

    def calculateKNNgraphDistanceMatrixPairwise(featureMatrix, para):
        r"""
        KNNgraphPairwise:  measuareName:k
        Pairwise:5
        Minkowski-Pairwise:5:1
        """
        measureName = ''
        k = 5
        if para != None:
            parawords = para.split(':')
            measureName = parawords[0]

        distMat = None
        if measureName == 'Pairwise':
            distMat = distance_matrix(featureMatrix, featureMatrix)
            k = int(parawords[1])
        elif measureName == 'Minkowski-Pairwise':
            p = int(parawords[2])
            distMat = minkowski_distance(featureMatrix, featureMatrix, p=p)
            k = int(parawords[1])
        else:
            print('meausreName in KNNgraph does not recongnized')
        edgeList = []

        for i in np.arange(distMat.shape[0]):
            res = distMat[:, i].argsort()[:k]
            for j in np.arange(k):
                edgeList.append((i, res[j]))

        return edgeList

    def calculateKNNgraphDistanceMatrix(featureMatrix, distanceType='euclidean', k=10):
        r"""
        KNNgraph:
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.cdist.html#scipy.spatial.distance.cdist
        distanceType incude:
        Distance functions between two numeric vectors u and v. Computing distances over a large collection of vectors is inefficient for these functions. Use pdist for this purpose.
        braycurtis(u, v[, w])	Compute the Bray-Curtis distance between two 1-D arrays.
        canberra(u, v[, w])	Compute the Canberra distance between two 1-D arrays.
        chebyshev(u, v[, w])	Compute the Chebyshev distance.
        cityblock(u, v[, w])	Compute the City Block (Manhattan) distance.
        correlation(u, v[, w, centered])	Compute the correlation distance between two 1-D arrays.
        cosine(u, v[, w])	Compute the Cosine distance between 1-D arrays.
        euclidean(u, v[, w])	Computes the Euclidean distance between two 1-D arrays.
        jensenshannon(p, q[, base])	Compute the Jensen-Shannon distance (metric) between two 1-D probability arrays.
        mahalanobis(u, v, VI)	Compute the Mahalanobis distance between two 1-D arrays.
        minkowski(u, v[, p, w])	Compute the Minkowski distance between two 1-D arrays.
        seuclidean(u, v, V)	Return the standardized Euclidean distance between two 1-D arrays.
        sqeuclidean(u, v[, w])	Compute the squared Euclidean distance between two 1-D arrays.
        wminkowski(u, v, p, w)	Compute the weighted Minkowski distance between two 1-D arrays.
        Distance functions between two boolean vectors (representing sets) u and v. As in the case of numerical vectors, pdist is more efficient for computing the distances between all pairs.
        dice(u, v[, w])	Compute the Dice dissimilarity between two boolean 1-D arrays.
        hamming(u, v[, w])	Compute the Hamming distance between two 1-D arrays.
        jaccard(u, v[, w])	Compute the Jaccard-Needham dissimilarity between two boolean 1-D arrays.
        kulsinski(u, v[, w])	Compute the Kulsinski dissimilarity between two boolean 1-D arrays.
        rogerstanimoto(u, v[, w])	Compute the Rogers-Tanimoto dissimilarity between two boolean 1-D arrays.
        russellrao(u, v[, w])	Compute the Russell-Rao dissimilarity between two boolean 1-D arrays.
        sokalmichener(u, v[, w])	Compute the Sokal-Michener dissimilarity between two boolean 1-D arrays.
        sokalsneath(u, v[, w])	Compute the Sokal-Sneath dissimilarity between two boolean 1-D arrays.
        yule(u, v[, w])	Compute the Yule dissimilarity between two boolean 1-D arrays.
        hamming also operates over discrete numerical vectors.
        """

        distMat = distance.cdist(featureMatrix, featureMatrix, distanceType)

        edgeList = []

        for i in np.arange(distMat.shape[0]):
            res = distMat[:, i].argsort()[:k]
            for j in np.arange(k):
                edgeList.append((i, res[j]))

        return edgeList

    edgeList = None
    adj = None

    def calculateThresholdgraphDistanceMatrix(featureMatrix, distanceType='euclidean', threshold=0.5):
        r"""
        Thresholdgraph: Graph with certain threshold
        """

        distMat = distance.cdist(featureMatrix, featureMatrix, distanceType)

        edgeList = []

        for i in np.arange(distMat.shape[0]):
            indexArray = np.where(distMat[i, :] > threshold)
            for j in indexArray[0]:
                edgeList.append((i, j))

        return edgeList

    def calculateKNNThresholdgraphDistanceMatrix(featureMatrix, distanceType='cosine', k=10, threshold=0.5):
        r"""
        Thresholdgraph: KNN Graph with certain threshold
        """

        distMat = distance.cdist(featureMatrix, featureMatrix, distanceType)

        edgeList = []

        for i in np.arange(distMat.shape[0]):
            res = distMat[:, i].argsort()[:k]
            for j in np.arange(k - 1):
                if (distMat[i, res[j]] > threshold):
                    edgeList.append((i, res[j]))
            # edgeList.append((i,res[k-1]))

        return edgeList

    def calculateKNNgraphDistanceMatrixML(featureMatrix, distanceType='euclidean', k=10, param=None):
        r"""
        Thresholdgraph: KNN Graph with Machine Learning based methods
        IsolationForest
        https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.IsolationForest.html#sklearn.ensemble.IsolationForest
        """

        distMat = distance.cdist(featureMatrix, featureMatrix, distanceType)
        edgeList = []

        # parallel: n_jobs=-1 for using all processors
        clf = IsolationForest(behaviour='new', contamination='auto', n_jobs=-1)

        for i in np.arange(distMat.shape[0]):
            res = distMat[i, :].argsort()[:k + 1]
            preds = clf.fit_predict(featureMatrix[res, :])
            for j in np.arange(1, k + 1):
                # weight = 1.0
                if preds[j] == -1:
                    weight = 0.0
                else:
                    weight = 1.0
                # preds[j]==-1 means outliner, 1 is what we want
                edgeList.append((i, res[j], weight))

        return edgeList

    def calculateKNNgraphDistanceMatrixStatsSingleThread(featureMatrix, distanceType='euclidean', k=10, param=None):
        r"""
        Thresholdgraph: KNN Graph with stats one-std based methods, SingleThread version
        """

        edgeList = []
        # Version 1: cost memory, precalculate all dist

        ## distMat = distance.cdist(featureMatrix,featureMatrix, distanceType)
        ## parallel
        # distMat = pairwise_distances(featureMatrix,featureMatrix, distanceType, n_jobs=-1)

        # for i in np.arange(distMat.shape[0]):
        #     res = distMat[:,i].argsort()[:k+1]
        #     tmpdist = distMat[res[1:k+1],i]
        #     mean = np.mean(tmpdist)
        #     std = np.std(tmpdist)
        #     for j in np.arange(1,k+1):
        #         if (distMat[i,res[j]]<=mean+std) and (distMat[i,res[j]]>=mean-std):
        #             weight = 1.0
        #         else:
        #             weight = 0.0
        #         edgeList.append((i,res[j],weight))

        ## Version 2: for each of the cell, calculate dist, save memory
        p_time = time.time()
        for i in np.arange(featureMatrix.shape[0]):
            if i % 10000 == 0:
                print('Start pruning ' + str(i) + 'th cell, cost ' + str(time.time() - p_time) + 's')
            tmp = featureMatrix[i, :].reshape(1, -1)
            distMat = distance.cdist(tmp, featureMatrix, distanceType)
            res = distMat.argsort()[:k + 1]
            tmpdist = distMat[0, res[0][1:k + 1]]
            boundary = np.mean(tmpdist) + np.std(tmpdist)
            for j in np.arange(1, k + 1):
                # should check, only exclude large outliners
                # if (distMat[0,res[0][j]]<=mean+std) and (distMat[0,res[0][j]]>=mean-std):
                if distMat[0, res[0][j]] <= boundary:
                    weight = 1.0
                else:
                    weight = 0.0
                edgeList.append((i, res[0][j], weight))
        return edgeList

    def calculateKNNgraphDistanceMatrixStats(featureMatrix, distanceType='euclidean', k=10, param=None,
                                             parallelLimit=0):
        r"""
        Thresholdgraph: KNN Graph with stats one-std based methods using parallel cores
        """
        edgeList = []
        # Get number of availble cores
        USE_CORES = 0
        NUM_CORES = multiprocessing.cpu_count()
        # if no limit, use all cores
        if parallelLimit == 0:
            USE_CORES = NUM_CORES
        # if limit < cores, use limit number
        elif parallelLimit < NUM_CORES:
            USE_CORES = parallelLimit
        # if limit is not valid, use all cores
        else:
            USE_CORES = NUM_CORES
        print('Start Pruning using ' + str(USE_CORES) + ' of ' + str(NUM_CORES) + ' available cores')

        t = time.time()
        # Use number of cpus for top-K finding
        with Pool(USE_CORES) as p:
            # edgeListT = p.map(vecfindK, range(featureMatrix.shape[0]))
            edgeListT = FindKParallel(featureMatrix, distanceType, k).work()

        t1 = time.time()
        print('Pruning succeed in ' + str(t1 - t) + ' seconds')
        flatten = lambda l: [item for sublist in l for item in sublist]
        t2 = time.time()
        edgeList = flatten(edgeListT)
        print('Prune out ready in ' + str(t2 - t1) + ' seconds')

        return edgeList

    class FindKParallel():
        '''
        A class to find K parallel
        '''

        def __init__(self, featureMatrix, distanceType, k):
            self.featureMatrix = featureMatrix
            self.distanceType = distanceType
            self.k = k

        def vecfindK(self, i):
            '''
            Find topK in paral
            '''
            edgeList_t = []
            # print('*'+str(i))
            tmp = self.featureMatrix[i, :].reshape(1, -1)
            distMat = distance.cdist(tmp, self.featureMatrix, self.distanceType)
            # print('#'+str(distMat))
            res = distMat.argsort()[:self.k + 1]
            # print('!'+str(res))
            tmpdist = distMat[0, res[0][1:self.k + 1]]
            # print('@'+str(tmpdist))
            boundary = np.mean(tmpdist) + np.std(tmpdist)
            # print('&'+str(boundary))
            for j in np.arange(1, self.k + 1):
                # check, only exclude large outliners
                # if (distMat[0,res[0][j]]<=mean+std) and (distMat[0,res[0][j]]>=mean-std):
                if distMat[0, res[0][j]] <= boundary:
                    weight = kernelDistance(distMat[0, res[0][j]])
                    edgeList_t.append((i, res[0][j], weight))
            # print('%'+str(len(edgeList_t)))
            return edgeList_t

        def work(self):
            return Pool().map(self.vecfindK, range(self.featureMatrix.shape[0]))

    def edgeList2edgeDict(edgeList, nodesize):
        graphdict = {}
        tdict = {}

        for edge in edgeList:
            end1 = edge[0]
            end2 = edge[1]
            tdict[end1] = ""
            tdict[end2] = ""
            if end1 in graphdict:
                tmplist = graphdict[end1]
            else:
                tmplist = []
            tmplist.append(end2)
            graphdict[end1] = tmplist

        # check and get full matrix
        for i in range(nodesize):
            if i not in tdict:
                graphdict[i] = []

        return graphdict

    if graphType == 'KNNgraphPairwise':
        edgeList = calculateKNNgraphDistanceMatrixPairwise(featureMatrix, para)
    elif graphType == 'KNNgraph':
        if para != None:
            parawords = para.split(':')
            distanceType = parawords[0]
            k = int(parawords[1])
        edgeList = calculateKNNgraphDistanceMatrix(featureMatrix, distanceType=distanceType, k=k)
    elif graphType == 'Thresholdgraph':
        if para != None:
            parawords = para.split(':')
            distanceType = parawords[0]
            threshold = float(parawords[1])
        edgeList = calculateThresholdgraphDistanceMatrix(featureMatrix, distanceType=distanceType, threshold=threshold)
    elif graphType == 'KNNgraphThreshold':
        if para != None:
            parawords = para.split(':')
            distanceType = parawords[0]
            k = int(parawords[1])
            threshold = float(parawords[2])
        edgeList = calculateKNNThresholdgraphDistanceMatrix(featureMatrix, distanceType=distanceType, k=k, threshold=threshold)
    elif graphType == 'KNNgraphML':
        # with weights!
        # https://towardsdatascience.com/5-ways-to-detect-outliers-that-every-data-scientist-should-know-python-code-70a54335a623
        # https://scikit-learn.org/stable/modules/outlier_detection.html
        if para != None:
            parawords = para.split(':')
            distanceType = parawords[0]
            k = int(parawords[1])
        edgeList = calculateKNNgraphDistanceMatrixML(featureMatrix, distanceType=distanceType, k=k)
    elif graphType == 'KNNgraphStats':
        # with weights!
        # with stats, one std is contained
        if para != None:
            parawords = para.split(':')
            distanceType = parawords[0]
            k = int(parawords[1])
        edgeList = calculateKNNgraphDistanceMatrixStats(featureMatrix, distanceType=distanceType, k=k, parallelLimit=parallelLimit)
    elif graphType == 'KNNgraphStatsSingleThread':
        # with weights!
        # with stats, one std is contained, but only use single thread
        if para != None:
            parawords = para.split(':')
            distanceType = parawords[0]
            k = int(parawords[1])
        edgeList = calculateKNNgraphDistanceMatrixStatsSingleThread(featureMatrix, distanceType=distanceType, k=k)
    else:
        print('Should give graphtype')

    if adjTag:
        graphdict = edgeList2edgeDict(edgeList, featureMatrix.shape[0])
        adj = nx.adjacency_matrix(nx.from_dict_of_lists(graphdict))

    return adj, edgeList


############################
#         stagate          #
############################

def stagate_construct_graph(adata,  rad_cutoff=None, k_cutoff=None, model='Radius', verbose=True):
    x_pixel = pd.DataFrame(adata.obs['x_pixel'])
    y_pixel = pd.DataFrame(adata.obs['y_pixel'])
    coor = pd.concat([y_pixel, x_pixel], axis=1)
    coor.index = adata.obs.index
    coor.columns = ['imagerow', 'imagecol']
    if model == 'Radius':
        nbrs = sklearn.neighbors.NearestNeighbors(radius=rad_cutoff).fit(coor)
        distances, indices = nbrs.radius_neighbors(coor, return_distance=True)
        KNN_list = []
        for it in range(indices.shape[0]):
            KNN_list.append(pd.DataFrame(zip([it]*indices[it].shape[0], indices[it], distances[it])))

    if model == 'KNN':
        nbrs = sklearn.neighbors.NearestNeighbors(n_neighbors=k_cutoff+1).fit(coor)
        distances, indices = nbrs.kneighbors(coor)
        KNN_list = []
        for it in range(indices.shape[0]):
            KNN_list.append(pd.DataFrame(zip([it]*indices.shape[1],indices[it,:], distances[it,:])))

    KNN_df = pd.concat(KNN_list)
    KNN_df.columns = ['Cell1', 'Cell2', 'Distance']

    Spatial_Net = KNN_df.copy()
    Spatial_Net = Spatial_Net.loc[Spatial_Net['Distance']>0,]
    id_cell_trans = dict(zip(range(coor.shape[0]), np.array(coor.index), ))
    Spatial_Net['Cell1'] = Spatial_Net['Cell1'].map(id_cell_trans)
    Spatial_Net['Cell2'] = Spatial_Net['Cell2'].map(id_cell_trans)
    if verbose:
        print('The graph contains %d edges, %d cells.' %(Spatial_Net.shape[0], adata.n_obs))
        print('%.4f neighbors per cell on average.' %(Spatial_Net.shape[0]/adata.n_obs))

    adata.uns['Spatial_Net'] = Spatial_Net
    
    G_df = adata.uns['Spatial_Net'].copy()
    cells = np.array(adata.obs_names)
    cells_id_tran = dict(zip(cells, range(cells.shape[0])))
    G_df['Cell1'] = G_df['Cell1'].map(cells_id_tran)
    G_df['Cell2'] = G_df['Cell2'].map(cells_id_tran)

    G = sp.coo_matrix((np.ones(G_df.shape[0]), (G_df['Cell1'], G_df['Cell2'])), shape=(adata.n_obs, adata.n_obs))
    G = G + sp.eye(G.shape[0])

    edgeList = np.nonzero(G)   
    return edgeList


# Copyright 2022 DSE lab.  All rights reserved.

# TODO：start from project root directory to solve the package import error
import os, sys

from sklearn.decomposition import PCA 

sys.path.append("../../..")

from dance.datasets.spatial import SpotDataset
from dance.modules.spatial.spatial_domain import louvain
from dance.modules.spatial.spatial_domain import stlearn
from dance.modules.spatial.spatial_domain.spagcn import SpaGCN
from dance.modules.spatial.spatial_domain.stlearn import stKmeans, stPrepare
from dance.transforms.graph_construct import construct_graph, neighbors_get_adj
from dance.transforms.preprocess import SME_normalize, extract_feature, normalize_total, pca, prefilter_cells, prefilter_genes, prefilter_specialgenes, normalize, log1p, sc_scale, scale , set_seed
import argparse
import scanpy as sc

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--sample_number", type=str, default="151673", \
        help="12 samples of human dorsolateral prefrontal cortex dataset \
        supported in the task of spatial domain task.")
    parser.add_argument("--n_clusters", type=int, default=17, help="the number of clusters")
    parser.add_argument("--n_components", type=int, default=50, help="the number of components in PCA")
    parser.add_argument("--device", type=str, default='cuda', help="device for resnet extract feature")
    args = parser.parse_args()

    # get data 
    dataset = SpotDataset(args.sample_number, data_dir="../../../data/spot")
    ## dataset.data has repeat name , be careful

    # build graph 
    # dataset.adj = construct_graph(dataset.data.obs['x'], dataset.data.obs['y'], dataset.data.obs['x_pixel'], dataset.data.obs['y_pixel'], dataset.img, beta=49, alpha=1, histology=True)

    # preprocess data 
    dataset.data.var_names_make_unique()

    # prefilter_cells(dataset.data) # this operation will change the data shape
    # prefilter_specialgenes(dataset.data)
    sc.pp.filter_genes(dataset.data,min_cells=1)
    normalize_total(dataset.data)
    log1p(dataset.data)

    extract_feature(dataset.data, image=dataset.img, n_components=args.n_components, device="cuda")
    # pca for gene expression data
    pca(dataset.data, n_components=args.n_components)

    # X_pca
    stPrepare(dataset.data)
    data_SME = dataset.data.copy()
    SME_normalize(data_SME, use_data='raw')
    data_SME.X = data_SME.obsm['raw_SME_normalized']
    sc_scale(data_SME)
 
    mypca = PCA(n_components=50)
    data_SME.obsm["X_pca"] = mypca.fit_transform(data_SME.X)
    
    # Option 1: kmeans
    # stKmeans(dataset.data, n_clusters=19)
    # model = stlearn.StKmeans(n_clusters=7)
    # stKmeans(dataset.data, n_clusters=args.n_clusters)
    # model = stlearn.StKmeans(n_clusters=args.n_clusters)
    # model.fit(dataset.data)
    # prediction = model.predict()
    # print("prediction:", prediction)
    # print(model.score(dataset.data.obs['label'].values)) # 0.1989

    # Option 2: Louvain
    data_SME.adj = neighbors_get_adj(data_SME, n_neighbors=args.n_clusters, use_rep='X_pca')
    model = stlearn.StLouvain()
    model.fit(data_SME, data_SME.adj, resolution=0.6)
    predict = model.predict()
    print(model.score(data_SME.obs['label'].values)) # 0. 31  # 0.366

""" To reproduce stlearn on other samples, please refer to command lines belows:
NOTE: since the stlearn method is unstable, you have to run multiple times to get 
      best performance. 
      
human dorsolateral prefrontal cortex sample 151676:
python stlearn.py --n_clusters=20 --sample_number=151673

human dorsolateral prefrontal cortex sample 151507:
python stlearn.py --n_clusters=20 --sample_number=151507
"""






"""
Trainer class for model training.
This code is based on https://github.com/NVlabs/MUNIT.
"""
from torch.autograd import Variable
import torch
import torch.nn as nn
import os
import torch.nn.init as init
import math
from torch.optim import lr_scheduler
import torch.nn.functional as F
from dance.utils import SimpleIndexDataset
from torch.utils.data import DataLoader
import sys

class Discriminator(nn.Module):

    def __init__(self, input_dim, params):
        super(Discriminator, self).__init__()
        self.gan_type = params['gan_type']
        self.dim = params['dim']
        self.norm = params['norm']
        self.input_dim = input_dim
        self.net = self._make_net()

    def _make_net(self):
        return nn.Sequential(
            nn.Linear(self.input_dim, self.input_dim),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Linear(self.input_dim, self.dim),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Linear(self.dim, 1),
            nn.LeakyReLU(0.2, inplace=True),
        )

    def forward(self, x):
        return self.net(x)

    def calc_dis_loss(self, input_fake, input_real):
        # calculate the loss to train D
        outs0 = [self.forward(input_fake)]
        outs1 = [self.forward(input_real)]
        loss = 0

        for it, (out0, out1) in enumerate(zip(outs0, outs1)):
            loss += torch.mean((out0 - 0) ** 2) + torch.mean((out1 - 1) ** 2)
        return loss

    def calc_gen_loss(self, input_fake):
        # calculate the loss to train G
        outs0 = [self.forward(input_fake)]
        loss = 0
        for it, (out0) in enumerate(outs0):
            # 1 = real data
            loss += torch.mean((out0 - 1) ** 2)
        return loss

    def calc_gen_loss_reverse(self, input_real):
        # calculate the loss to train G
        outs0 = [self.forward(input_real)]
        loss = 0
        for it, (out0) in enumerate(outs0):
            # 0 = fake data
            loss += torch.mean((out0 - 0) ** 2)
        return loss

    def calc_gen_loss_half(self, input_fake):
        # calculate the loss to train G
        outs0 = [self.forward(input_fake)]
        loss = 0
        for it, (out0) in enumerate(outs0):
            if self.gan_type == 'lsgan':
                loss += torch.mean((out0 - 0.5) ** 2)
            elif self.gan_type == 'nsgan':
                all1 = Variable(torch.ones_like(out0.data).cuda(), requires_grad=False)
                loss += torch.mean(F.binary_cross_entropy(F.sigmoid(out0), all1))
            else:
                assert 0, "Unsupported GAN type: {}".format(self.gan_type)
        return loss


##################################################################################
# Generator
##################################################################################

class VAEGen(nn.Module):
    # VAE architecture
    def __init__(self, input_dim, params, shared_layer=False):
        super(VAEGen, self).__init__()
        self.dim = params['dim']
        self.latent = params['latent']
        self.input_dim = input_dim

        # encoder_layers = [nn.Linear(self.input_dim, self.input_dim),
        #                   nn.LeakyReLU(0.2, inplace=True),
        #                   nn.Linear(self.input_dim, self.input_dim),
        #                   nn.LeakyReLU(0.2, inplace=True),
        #                   nn.Linear(self.input_dim, self.input_dim),
        #                   nn.LeakyReLU(0.2, inplace=True),
        #                   nn.Linear(self.input_dim, self.dim),
        #                   nn.LeakyReLU(0.2, inplace=True)]
        #
        # decoder_layers = [nn.LeakyReLU(0.2, inplace=True),
        #                   nn.Linear(self.dim, self.input_dim),
        #                   nn.LeakyReLU(0.2, inplace=True),
        #                   nn.Linear(self.input_dim, self.input_dim),
        #                   nn.LeakyReLU(0.2, inplace=True),
        #                   nn.Linear(self.input_dim, self.input_dim),
        #                   nn.LeakyReLU(0.2, inplace=True),
        #                   nn.Linear(self.input_dim, self.input_dim),
        #                   nn.LeakyReLU(0.2, inplace=True)]

        if self.input_dim>1000:
            hid_size = 1000
        else:
            hid_size = self.input_dim

        encoder_layers = [nn.Linear(self.input_dim, hid_size),
                          nn.LeakyReLU(0.2, inplace=True),
                          nn.Linear(hid_size, hid_size),
                          nn.LeakyReLU(0.2, inplace=True),
                          nn.Linear(hid_size, hid_size),
                          nn.LeakyReLU(0.2, inplace=True),
                          nn.Linear(hid_size, self.dim),
                          nn.LeakyReLU(0.2, inplace=True)]

        decoder_layers = [nn.LeakyReLU(0.2, inplace=True),
                          nn.Linear(self.dim, hid_size),
                          nn.LeakyReLU(0.2, inplace=True),
                          nn.Linear(hid_size, hid_size),
                          nn.LeakyReLU(0.2, inplace=True),
                          nn.Linear(hid_size, hid_size),
                          nn.LeakyReLU(0.2, inplace=True),
                          nn.Linear(hid_size, self.input_dim),
                          nn.LeakyReLU(0.2, inplace=True)]

        if shared_layer:
            encoder_layers += [shared_layer["enc"], nn.LeakyReLU(0.2, inplace=True)]
            decoder_layers = [shared_layer["dec"]] + decoder_layers
        else:
            encoder_layers += [nn.Linear(self.dim, self.latent), nn.LeakyReLU(0.2, inplace=True)]
            decoder_layers = [nn.Linear(self.latent, self.dim)] + decoder_layers
        self.enc = nn.Sequential(*encoder_layers)
        self.dec = nn.Sequential(*decoder_layers)

    def forward(self, images):
        # This is a reduced VAE implementation where we assume the outputs are multivariate Gaussian distribution with mean = hiddens and std_dev = all ones.
        hiddens = self.encode(images)
        if self.training == True:
            noise = Variable(torch.randn(hiddens.size()).cuda(hiddens.data.get_device()))
            images_recon = self.decode(hiddens + noise)
        else:
            images_recon = self.decode(hiddens)
        return images_recon, hiddens

    def encode(self, images):
        hiddens = self.enc(images)
        noise = Variable(torch.randn(hiddens.size()).cuda(hiddens.data.get_device()))
        return hiddens, noise

    def decode(self, hiddens):
        images = self.dec(hiddens)
        return images

##################################################################################
# Classifier
##################################################################################

class Classifier(nn.Module):
    def __init__(self, input_dim, cls=3):
        super(Classifier, self).__init__()
        self.input_dim = input_dim
        self.classes = cls
        self.net = self._make_net()

        self.cel = nn.CrossEntropyLoss()

    def _make_net(self):
        return nn.Sequential(
            nn.Linear(self.input_dim, self.classes)
        )

    def forward(self, x):
        return self.net(x)

    def class_loss(self, input, target):
        return self.cel(input, target)

def weights_init(init_type='gaussian'):
    def init_fun(m):
        classname = m.__class__.__name__
        if (classname.find('Conv') == 0 or classname.find('Linear') == 0) and hasattr(m, 'weight'):
            # print m.__class__.__name__
            if init_type == 'gaussian':
                init.normal_(m.weight.data, 0.0, 0.02)
            elif init_type == 'xavier':
                init.xavier_normal_(m.weight.data, gain=math.sqrt(2))
            elif init_type == 'kaiming':
                init.kaiming_normal_(m.weight.data, a=0, mode='fan_in')
            elif init_type == 'orthogonal':
                init.orthogonal_(m.weight.data, gain=math.sqrt(2))
            elif init_type == 'default':
                pass
            else:
                assert 0, "Unsupported initialization: {}".format(init_type)
            if hasattr(m, 'bias') and m.bias is not None:
                init.constant_(m.bias.data, 0.0)

    return init_fun

# Get model list for resume
def get_model_list(dirname, key):
    if os.path.exists(dirname) is False:
        return None
    gen_models = [os.path.join(dirname, f) for f in os.listdir(dirname) if
                  os.path.isfile(os.path.join(dirname, f)) and key in f and ".pt" in f]
    if gen_models is None:
        return None
    gen_models.sort()
    last_model_name = gen_models[-1]
    return last_model_name

def get_scheduler(optimizer, hyperparameters, iterations=-1):
    if 'lr_policy' not in hyperparameters or hyperparameters['lr_policy'] == 'constant':
        scheduler = None  # constant scheduler
    elif hyperparameters['lr_policy'] == 'step':
        scheduler = lr_scheduler.StepLR(optimizer, step_size=hyperparameters['step_size'],
                                        gamma=hyperparameters['gamma'], last_epoch=iterations)
    else:
        return NotImplementedError('learning rate policy [%s] is not implemented', hyperparameters['lr_policy'])
    return scheduler

class CMAE(nn.Module):
    """CMAE class.

    Parameters
    ----------
    hyperparameters : dictionary
        A dictionary that contains arguments of CMAE. For details of parameters in parser args, please refer to link (parser help document).
    """

    def __init__(self, hyperparameters):
        super(CMAE, self).__init__()
        lr = hyperparameters['lr']
        # Initiate the networks
        shared_layer = False
        if "shared_layer" in hyperparameters and hyperparameters["shared_layer"]:
            shared_layer = {}
            shared_layer["dec"] = nn.Linear(hyperparameters['gen']['latent'], hyperparameters['gen']['dim'])
            shared_layer["enc"] = nn.Linear(hyperparameters['gen']['dim'], hyperparameters['gen']['latent'])

        self.gen_a = VAEGen(hyperparameters['input_dim_a'], hyperparameters['gen'],
                            shared_layer)  # auto-encoder for domain a
        self.gen_b = VAEGen(hyperparameters['input_dim_b'], hyperparameters['gen'],
                            shared_layer)  # auto-encoder for domain b
        self.dis_latent = Discriminator(hyperparameters['gen']['latent'],
                                        hyperparameters['dis'])  # discriminator for latent space

        self.classifier = Classifier(hyperparameters['gen']['latent'], cls=hyperparameters['num_of_classes'])  # classifier on the latent space

        # Setup the optimizers
        beta1 = hyperparameters['beta1']
        beta2 = hyperparameters['beta2']
        dis_params = list(self.dis_latent.parameters())
        gen_params = list(self.gen_a.parameters()) + list(self.gen_b.parameters()) + list(self.classifier.parameters())
        self.dis_opt = torch.optim.AdamW([p for p in dis_params if p.requires_grad],
                                        lr=lr, betas=(beta1, beta2), weight_decay=hyperparameters['weight_decay'])
        self.gen_opt = torch.optim.AdamW([p for p in gen_params if p.requires_grad],
                                        lr=lr, betas=(beta1, beta2), weight_decay=hyperparameters['weight_decay'])
        self.dis_scheduler = get_scheduler(self.dis_opt, hyperparameters)
        self.gen_scheduler = get_scheduler(self.gen_opt, hyperparameters)

        # Network weight initialization
        self.apply(weights_init(hyperparameters['init']))
        self.dis_latent.apply(weights_init('gaussian'))
        self.hyperparameters = hyperparameters

    def _mae_loss(self, input, target):
        """A simple criterion function for MAE loss.

        Parameters
        ----------
        inputs : torch.Tensor
            A input tensor.
        target : torch.Tensor
            A target tensor.

        Returns
        -------
        loss : float
            MAE loss between input and target tensors.

        """
        return torch.mean(torch.abs(input - target))

    def predict(self, mod1):
        """Predict function to get prediction of target modality features.

        Parameters
        ----------
        mod1 : torch.Tensor
            Input modality features.

        Returns
        -------
        pred : torch.Tensor
            Predicted features of target modality.

        """
        with torch.no_grad():
            emb, _ = self.gen_a.encode(mod1)
            pred = self.gen_b.decode(emb)
        return pred

    def score(self, mod1, mod2):
        """Score function to get score of prediction.

        Parameters
        ----------
        mod1 : torch.Tensor
            Input modality features.
        mod2 : torch.Tensor
            Output modality features.

        Returns
        -------
        score : float
            RMSE loss of predicted output modality features.

        """
        with torch.no_grad():
            pred = self.predict(mod1)
            mse = nn.MSELoss()
            score = math.sqrt(mse(pred, mod2))
            return score

    def forward(self, mod1, mod2):
        """Forward function for torch.nn.Module.

        Parameters
        ----------
        mod1 : torch.Tensor
            Input modality features.
        mod2 : torch.Tensor
            Target modality features.

        Returns
        -------
        x_ab : torch.Tensor
            Prediction of target modality from input modality.
        x_ba : torch.Tensor
            Prediction of input modality from target modality.
        """
        self.eval()
        h_a, _ = self.gen_a.encode(mod1)
        h_b, _ = self.gen_b.encode(mod2)
        x_ba = self.gen_a.decode(h_b)
        x_ab = self.gen_b.decode(h_a)
        self.train()
        return x_ab, x_ba

    def _gen_update(self, x_a, x_b, super_a, super_b, hyperparameters, a_labels=None, b_labels=None, variational=True):
        true_samples = Variable(
            torch.randn(200, hyperparameters['gen']['latent']),
            requires_grad=False
        ).cuda()

        self.gen_opt.zero_grad()
        # encode
        h_a, n_a = self.gen_a.encode(x_a)
        h_b, n_b = self.gen_b.encode(x_b)
        # decode (within domain)
        if variational:
            h_a = h_a + n_a
            h_b = h_b + n_b

        x_a_recon = self.gen_a.decode(h_a)
        x_b_recon = self.gen_b.decode(h_b)

        # decode (cross domain)
        x_ba = self.gen_a.decode(h_b)
        x_ab = self.gen_b.decode(h_a)
        # encode again
        h_b_recon, n_b_recon = self.gen_a.encode(x_ba)
        h_a_recon, n_a_recon = self.gen_b.encode(x_ab)
        # decode again (if needed)
        if variational:
            h_a_recon = h_a_recon + n_a_recon
            h_b_recon = h_b_recon + n_b_recon

        classes_a = self.classifier.forward(h_a)
        classes_b = self.classifier.forward(h_b)

        # reconstruction loss
        self.loss_gen_recon_x_a = self._mae_loss(x_a_recon, x_a)
        self.loss_gen_recon_x_b = self._mae_loss(x_b_recon, x_b)

        # GAN loss
        self.loss_latent_a = self.dis_latent.calc_gen_loss(h_a)
        self.loss_latent_b = self.dis_latent.calc_gen_loss_reverse(h_b)

        # Classification Loss
        if a_labels is not None and b_labels is not None:
            self.loss_class_a = self.classifier.class_loss(classes_a, a_labels)
            self.loss_class_b = self.classifier.class_loss(classes_b, b_labels)
        else:
            self.loss_class_a = self.loss_class_b = 0

        # supervision
        s_a, n_a = self.gen_a.encode(super_a)
        s_b, n_b = self.gen_b.encode(super_b)

        self.loss_supervision = self._mae_loss(s_a, s_b)

        class_weight = hyperparameters['gan_w'] if "class_w" not in hyperparameters else hyperparameters["class_w"]

        # total loss
        self.loss_gen_total = hyperparameters['gan_w'] * self.loss_latent_a + \
                              hyperparameters['gan_w'] * self.loss_latent_b + \
                              class_weight * self.loss_class_a + \
                              class_weight * self.loss_class_b + \
                              hyperparameters['recon_x_w'] * self.loss_gen_recon_x_a + \
                              hyperparameters['recon_x_w'] * self.loss_gen_recon_x_b + \
                              hyperparameters['super_w'] * self.loss_supervision

        if variational:
            self.loss_gen_total += hyperparameters['recon_kl_w'] * self.loss_gen_recon_kl_a + \
                                   hyperparameters['recon_kl_w'] * self.loss_gen_recon_kl_b

        self.loss_gen_total.backward()
        self.gen_opt.step()

    def _sample(self, x_a, x_b):
        self.eval()
        x_a_recon, x_b_recon, x_ba, x_ab = [], [], [], []
        for i in range(x_a.size(0)):
            h_a, _ = self.gen_a.encode(x_a[i].unsqueeze(0))
            h_b, _ = self.gen_b.encode(x_b[i].unsqueeze(0))
            x_a_recon.append(self.gen_a.decode(h_a))
            x_b_recon.append(self.gen_b.decode(h_b))
            x_ba.append(self.gen_a.decode(h_b))
            x_ab.append(self.gen_b.decode(h_a))
        x_a_recon, x_b_recon = torch.cat(x_a_recon), torch.cat(x_b_recon)
        x_ba = torch.cat(x_ba)
        x_ab = torch.cat(x_ab)
        self.train()
        return x_a, x_a_recon, x_ab, x_b, x_b_recon, x_ba

    def _dis_update(self, x_a, x_b, hyperparameters):
        self.dis_opt.zero_grad()
        # encode
        h_a, n_a = self.gen_a.encode(x_a)
        h_b, n_b = self.gen_b.encode(x_b)
        # D loss
        self.loss_dis_latent = self.dis_latent.calc_dis_loss(h_a, h_b)
        self.loss_dis_total = hyperparameters['gan_w'] * (self.loss_dis_latent)
        self.loss_dis_total.backward()
        self.dis_opt.step()

    def _update_learning_rate(self):
        if self.dis_scheduler is not None:
            self.dis_scheduler.step()
        if self.gen_scheduler is not None:
            self.gen_scheduler.step()

    def resume(self, checkpoint_dir):
        """Resume function to resume from checkpoint file.

        Parameters
        ----------
        checkpoint_dir : str
            Path to the checkpoint file.

        Returns
        -------
        iterations : int
            Current iteration number of resumed model.

        """

        hyperparameters = self.hyperparameters
        # Load generators
        last_model_name = get_model_list(checkpoint_dir, "gen")
        state_dict = torch.load(last_model_name)
        self.gen_a.load_state_dict(state_dict['a'])
        self.gen_b.load_state_dict(state_dict['b'])
        iterations = int(last_model_name[-11:-3])
        # Load discriminators
        last_model_name = get_model_list(checkpoint_dir, "dis")
        state_dict = torch.load(last_model_name)
        self.dis_latent.load_state_dict(state_dict['latent'])
        # Load optimizers
        state_dict = torch.load(os.path.join(checkpoint_dir, 'optimizer.pt'))
        self.dis_opt.load_state_dict(state_dict['dis'])
        self.gen_opt.load_state_dict(state_dict['gen'])
        # Reinitilize schedulers
        self.dis_scheduler = get_scheduler(self.dis_opt, hyperparameters, iterations)
        self.gen_scheduler = get_scheduler(self.gen_opt, hyperparameters, iterations)
        print('Resume from iteration %d' % iterations)
        return iterations

    def save(self, checkpoint_dir, iterations):
        """Save function to save parameters to checkpoint file.

        Parameters
        ----------
        checkpoint_dir : str
            Path to the checkpoint file.
        iterations : int
            Current number of training iterations.

        Returns
        -------
        None.

        """
        # Save generators, discriminators, and optimizers
        gen_name = os.path.join(checkpoint_dir, 'gen_%08d.pt' % (iterations + 1))
        dis_name = os.path.join(checkpoint_dir, 'dis_%08d.pt' % (iterations + 1))
        opt_name = os.path.join(checkpoint_dir, 'optimizer.pt')
        torch.save(
            {'a': self.gen_a.state_dict(), 'b': self.gen_b.state_dict(), "classifier": self.classifier.state_dict()},
            gen_name)
        torch.save({'latent': self.dis_latent.state_dict()}, dis_name)
        torch.save({'gen': self.gen_opt.state_dict(), 'dis': self.dis_opt.state_dict()}, opt_name)

    def fit(self, train_mod1, train_mod2, aux_labels = None, checkpoint_directory='./checkpoint'):
        """Train CMAE.

        Parameters
        ----------
        train_mod1 : torch.Tensor
            Features of input modality.
        train_mod2 : torch.Tensor
            Features of target modality.
        aux_labels : torch.Tensor optional
            Auxiliary labels for extra supervision during training.
        checkpoint_directory : str optional
            Path to the checkpoint file, by default to be './checkpoint'.

        Returns
        -------
        None.

        """

        hyperparameters = self.hyperparameters
        idx = torch.randperm(train_mod1.shape[0])
        train_idx = idx[:int(idx.shape[0]*0.85)]
        val_idx = idx[int(idx.shape[0]*0.85):]

        train_dataset = SimpleIndexDataset(train_idx)
        train_loader = DataLoader(
            dataset=train_dataset,
            batch_size=hyperparameters['batch_size'],
            shuffle=True,
            num_workers=0,
            drop_last=True,
        )

        # Start training
        iterations = self.resume(checkpoint_directory, hyperparameters=hyperparameters) if hyperparameters['resume'] else 0
        num_disc = 1 if "num_disc" not in hyperparameters else hyperparameters["num_disc"]
        num_gen = 1 if "num_gen" not in hyperparameters else hyperparameters["num_gen"]

        while True:
            print('Iteration: ', iterations)
            for it, batch_idx in enumerate(train_loader):
                self._update_learning_rate()
                mod1, mod2 = train_mod1[batch_idx], train_mod2[batch_idx]

                for _ in range(num_disc):
                    self._dis_update(mod1, mod2, hyperparameters)
                for _ in range(num_gen):
                    if aux_labels is not None:
                        self._gen_update(mod1, mod2, mod1, mod2, hyperparameters, aux_labels[batch_idx], aux_labels[batch_idx],
                                         variational=False)
                    else:
                        self._gen_update(mod1, mod2, mod1, mod2, hyperparameters,
                                         variational=False)

            print('RMSE Loss:', self.score(train_mod1[val_idx], train_mod2[val_idx]))

            iterations += 1
            if iterations >= hyperparameters['max_iter']:
                print('Finish training')
                break